import { Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Model } from 'mongoose';
import { OrientationIndicator, OrientationIndicatorDocument } from '../schemas/orientation-indicator.schema';
import { User } from '../../users/schemas/user.schema';
import { IUser } from '../../users/interfaces/user.interface';
import { StructuresService } from '../../structures/services/structures.service';

@Injectable()
export class IndicatorService {
  private readonly logger = new Logger(IndicatorService.name);

  constructor(
    private structuresService: StructuresService,
    @InjectModel(OrientationIndicator.name) private OrientationIndicatorModel: Model<OrientationIndicatorDocument>,
    @InjectModel(User.name) private userModel: Model<IUser>
  ) {}

  public async createOrientation(orientationIndicator: OrientationIndicator): Promise<OrientationIndicator> {
    this.logger.debug('new orientation indicator');

    // Remove name (because the API is public)
    orientationIndicator.origin.nom = undefined;

    if (orientationIndicator.origin.adresse?.postcode && orientationIndicator.origin.adresse.commune) {
      const inseeCode = await this.structuresService.getInseeCode(
        orientationIndicator.origin.adresse?.postcode,
        orientationIndicator.origin.adresse?.commune
      );

      // Remove address of the origin and add inseeCode
      orientationIndicator.origin.adresse.numero = undefined;
      orientationIndicator.origin.adresse.commune = undefined;
      orientationIndicator.origin.adresse.postcode = undefined;
      orientationIndicator.origin.adresse['coordinates'] = undefined;
      orientationIndicator.origin.adresse.inseeCode = inseeCode;
    }

    return this.OrientationIndicatorModel.create(orientationIndicator);
  }

  public async getAllOrientation(
    limit?: number,
    skip?: number,
    startDate?: string,
    endDate?: string
  ): Promise<OrientationIndicator[]> {
    this.logger.debug('get all orientation indicators');

    if (!limit && !skip && !startDate && !endDate) {
      return this.OrientationIndicatorModel.find({});
    }

    const query = {};

    if (startDate || endDate) {
      query['createdAt'] = {};
      if (startDate) {
        query['createdAt']['$gte'] = new Date(startDate);
      }
      if (endDate) {
        query['createdAt']['$lte'] = new Date(endDate);
      }
    }

    if (isNaN(limit) || limit <= 0) {
      limit = 10;
    }

    if (isNaN(skip) || skip < 0) {
      skip = 0;
    }

    return this.OrientationIndicatorModel.find(query).limit(Number(limit)).skip(Number(skip));
  }

  public async countUsers(startDate?: string, endDate?: string): Promise<number> {
    this.logger.debug('count users');

    const conditions = [];

    if (startDate || endDate) {
      conditions.push({ lastLoginDate: { $exists: true } });

      if (startDate) {
        conditions.push({ lastLoginDate: { $gte: new Date(startDate) } });
      }
      if (endDate) {
        conditions.push({ lastLoginDate: { $lte: new Date(endDate) } });
      }
    }

    if (conditions.length > 0) {
      return this.userModel.countDocuments({ $and: conditions });
    } else {
      return this.userModel.countDocuments({});
    }
  }
}
