import { HttpService } from '@nestjs/axios';
import { Injectable, Logger } from '@nestjs/common';
import { AxiosResponse } from 'axios';
import * as FormData from 'form-data';
import * as fs from 'fs';
import * as path from 'path';
import { ConfigurationService } from '../configuration/configuration.service';
import { lastValueFrom } from 'rxjs';
@Injectable()
export class MailerService {
  public config = null;
  constructor(
    private readonly httpService: HttpService,
    private configurationService: ConfigurationService
  ) {
    this.config = this.configurationService.config;
  }

  public async send(to: string | { email: string }[], subject: string, html: string): Promise<AxiosResponse<unknown>> {
    const emailsToSend = typeof to === 'string' ? [{ email: to }] : to;
    const formData = new FormData();
    const data = JSON.stringify({
      from_email: process.env.MAIL_FROM,
      from_name: this.config.from_name,
      to: emailsToSend,
      reply_to: 'inclusionnumerique@grandlyon.com',
      subject: subject,
      content: this.addSignature(html),
    });
    formData.append('metadata', data);
    const contentLength = formData.getLengthSync();
    const headers = {
      'Content-Length': contentLength,
      Authorization: 'Bearer ' + process.env.MAIL_TOKEN,
      ...formData.getHeaders(),
    };
    if (process.env.MAIL_APIKEY) {
      headers['apikey'] = process.env.MAIL_APIKEY;
    }
    Logger.log(`Send mail to ${JSON.stringify(emailsToSend)} : ${subject}`, 'Mailer');
    return lastValueFrom(
      this.httpService.post(process.env.MAIL_URL, formData, {
        headers: headers,
      })
    ).then(
      (body) => {
        Logger.log(`Send mail to ${JSON.stringify(emailsToSend)} - success : ${subject}`, 'Mailer');
        return body.data;
      },
      (err) => {
        Logger.error(err, 'Mailer');
        throw err;
      }
    );
  }

  /**
   * Get email template location from config directory and given filename. Also, check if file exists.
   */
  public getTemplateLocation(filename: string) {
    const ejsPath = path.join(this.config.templates.directory, filename);
    if (!fs.existsSync(ejsPath)) {
      throw new Error(`Email template '${filename}' cannot be found in ${this.config.templates.directory}`);
    }
    return ejsPath;
  }

  /**
   * Load email json config file from config directory and given filename. Also, check if file exists
   */
  public loadJsonConfig(filename: string) {
    const jsonPath = path.join(this.config.templates.directory, filename);
    if (!fs.existsSync(jsonPath)) {
      throw new Error(`Email json definition file '${filename}' cannot be found in ${this.config.templates.directory}`);
    }
    return JSON.parse(fs.readFileSync(jsonPath).toString());
  }

  /**
   * Add site mail signature to a given html
   */
  public addSignature(html: string): string {
    html += `<br /><br /><p>L’équipe Rés'in</p><p style="font-family:helvetica;font-size:24px;font-weight:bold;margin:0;">rés<span style="color:red;font-weight:400;">’in</span></p><br /><br /><p>Ce message est automatique, merci de ne pas y répondre.<br/><br/>Pour toute question, nous sommes joignables à l'adresse email <a href="mailto:inclusionnumerique@grandlyon.com">inclusionnumerique@grandlyon.com</a>.</p>`;
    return html;
  }
}
