import { AnalysisAnalyzer, IndicesIndexSettings } from '@elastic/elasticsearch/lib/api/types';
import { ConfigurationService } from '../configuration/configuration.service';
import { Page } from '../pages/schemas/page.schema';
import { Post } from '../posts/schemas/post.schema';
import { UserRole } from '../users/enum/user-role.enum';
import { User } from '../users/schemas/user.schema';
import { connect } from 'puppeteer-core';
// eslint-disable-next-line @typescript-eslint/no-var-requires
const crypto = require('crypto');
export const md5 = (data: string): string => crypto.createHash('md5').update(data).digest('hex');

export function rewriteGhostImgUrl(configService: ConfigurationService, itemData: Page | Post): Page | Post {
  // Handle image display. Rewrite image URL to fit ghost infra issue.
  if (!configService.isLocalConf() && process.env.IS_OPENSHIFT !== 'true') {
    if (itemData.feature_image) {
      itemData.feature_image = `https://${configService.config.host}/blog/content${
        itemData.feature_image.split('/content')[1]
      }`;
    }
    const regex = /(https?:\/\/ghost):(\d*)/g;
    itemData.html = itemData.html.replace(regex, `https://${configService.config.host}/blog`);
  }
  return itemData;
}

export function hasAdminRole(user: User): boolean {
  return user?.role === UserRole.admin;
}

const analyzer: Record<string, AnalysisAnalyzer> = {
  homemade_french: {
    type: 'custom',
    tokenizer: 'standard',
    filter: ['lowercase', 'asciifolding', 'french_elision', 'french_stop', 'french_stemmer'],
  },
  homemade_french_stopless: {
    type: 'custom',
    tokenizer: 'standard',
    filter: ['lowercase', 'asciifolding', 'french_elision', 'french_stemmer'],
  },
};

export const es_settings_homemade_french: IndicesIndexSettings = {
  analysis: {
    filter: {
      french_elision: {
        type: 'elision',
        articles_case: true,
        articles: ['l', 'm', 't', 'qu', 'n', 's', 'j', 'd', 'c', 'jusqu', 'quoiqu', 'lorsqu', 'puisqu'],
      },
      french_stop: {
        type: 'stop',
        stopwords: '_french_',
      },
      french_stemmer: {
        type: 'stemmer',
        language: 'minimal_french',
      },
    },
    analyzer: analyzer,
  },
};

export function escapeElasticsearchQuery(query: string): string {
  return query.replace(/([+\-=><!(){}[\]^"~?:\\/]|&&|\|\|)/g, '\\$&');
}

/**
 *  Remove accents and trailing spaces from a string, replace non-alphanumeric characters with dashes and return the result in lowercase
 * @example
 * sanitize(' Métropole de Lyon ') => 'metropole-de-lyon'
 */
export const sanitize = (str: string) => {
  return str
    .trim()
    .toLowerCase()
    .normalize('NFD')
    .replace(/[\u0300-\u036f]/g, '') // remove accents
    .replace(/\s-\s/g, '-') // replace " - " by "-"
    .replace(/\s\/\s/g, '-') // replace " / " by "-"
    .replace(/\s/g, '-') // replace spaces by "-"
    .replace(/[^\w\s-]/g, ''); // remove all non alphanumeric characters except spaces and dashes
};

export async function generatePDF(url: string): Promise<Uint8Array> {
  const browser = await connect({
    // Default value for local dev
    browserWSEndpoint: process.env.CHROME_WS_URL || 'ws://localhost:5000',
  });

  const page = await browser.newPage();

  await page.goto(url, { waitUntil: 'networkidle0' }); // wait until all is loaded

  const buffer = await page.pdf({
    format: 'A4',
    printBackground: true, // Include background colors and images
    margin: { top: '20mm', right: '10mm', bottom: '20mm', left: '10mm' },
  });

  await browser.close();
  return buffer;
}
