import { Injectable, Logger } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { StructureDocument } from '../schemas/structure.schema';
import { Model } from 'mongoose';
import { toOsmOpeningHours } from '@gouvfr-anct/timetable-to-osm-opening-hours';
import { StructureCategories } from '../schemas/structureCategories.schema';
import * as config from './structures-export-configs/services-config.json';
import { StructureFormatted } from '../interfaces/structure-formatted.interface';
import { StructuresService } from './structures.service';

@Injectable()
export class StructuresExportService {
  private readonly logger = new Logger(StructuresExportService.name);

  constructor(
    private readonly structureService: StructuresService,
    @InjectModel('Structure') private structureModel: Model<StructureDocument>
  ) {}

  public async exportFormatted(): Promise<StructureFormatted[]> {
    try {
      this.logger.debug('Exporting formatted structures, only returning structures who consented to data sharing');
      const structures = await this.structureModel
        .find({ deletedAt: { $exists: false }, accountVerified: true, $where: 'this.dataShareConsentDate != null' })
        .populate({ path: 'personalOffers', model: 'PersonalOffer' })
        .select('-accountVerified -otherDescription -dataShareConsentDate')
        .exec();

      return structures.map((structure) => {
        // Set the offers from the structure and from its personalOffers
        this.structureService.setCategoriesWithPersonalOffers(structure, true);
        // Set the categories with personalOffers in structure.categories which is used in the services export below
        structure.categories.onlineProcedures = structure.categoriesWithPersonalOffers.onlineProcedures;
        structure.categories.baseSkills = structure.categoriesWithPersonalOffers.baseSkills;
        structure.categories.advancedSkills = structure.categoriesWithPersonalOffers.advancedSkills;
        const formatted: StructureFormatted = {
          // TODO add antenne field
          id: structure.id,
          adresse: this.formatAddress(structure.address),
          code_postal: structure.address?.postcode,
          commune: structure.address?.commune,
          date_maj: this.formatDate(structure.updatedAt),
          nom: structure.structureName,
          pivot: '00000000000000', // for now we don't have a pivot
          services: this.formatServices(structure.categories),
          source: 'Res-in',
          conditions_acces: this.checkAccessConditions(structure.categories.accessModality),
          ...(structure.address?.inseeCode && { code_insee: structure.address.inseeCode }),
          ...(structure.contactMail && { courriel: structure.contactMail }),
          ...(structure.coord && { longitude: structure.coord[0] }),
          ...(structure.coord && { latitude: structure.coord[1] }),
          ...(structure.description && { presentation_detail: structure.description }),
          ...((structure.categories.age ||
            structure.categories.handicaps ||
            structure.categories.languageAndIlliteracy ||
            structure.categories.genre) && {
            publics_accueillis: this.checkPublic(
              structure.categories.age,
              structure.categories.handicaps,
              structure.categories.languageAndIlliteracy,
              structure.categories.genre
            ),
          }),
          ...(structure.website && { site_web: structure.website }),
          ...(structure.contactPhone && { telephone: structure.contactPhone }),
          ...(structure.facebook && { facebook: structure.facebook }),
          ...(structure.twitter && { twitter: structure.twitter }),
          ...(structure.instagram && { instagram: structure.instagram }),
          ...(structure.linkedin && { linkedin: structure.linkedin }),
          ...(structure.freeWorkShop && { atelier_gratuit: structure.freeWorkShop }),
          ...(structure.nbPrinters && { nb_imprimantes: structure.nbPrinters }),
          ...(structure.nbComputers && { nb_ordinateurs: structure.nbComputers }),
          ...(structure.nbScanners && { nb_scanners: structure.nbScanners }),
          ...(structure.hours && { horaires: this.convertToOSMFormat(structure.hours) }),
          acces_pmr: structure.pmrAccess,
        };

        return formatted;
      });
    } catch (error) {
      this.logger.error('Error occurred while exporting formatted structures:', error);
      return [];
    }
  }

  private checkPublic(age: string[], handicaps: string[], languages: string[], other: string[]): string {
    if (!age && !handicaps && !other && !languages) {
      return '';
    }

    const ageMapping = {
      family: 'Familles/enfants',
      young: 'Jeunes (16-26 ans)',
      adults: 'Adultes',
      old: 'Seniors (+ 65 ans)',
    };

    const handicapsMapping = {
      blindness: 'Déficience visuelle',
      deafness: 'Surdité',
      mentalDisability: 'Handicaps mentaux : déficiences limitant les activités d’une personne',
    };

    const languageMapping = {
      illettrisme: 'Personnes en situation d’illettrisme',
      english: 'Public langues étrangères',
      foreignLanguage: 'Public langues étrangères',
    };

    const otherMapping = {
      uniquementFemmes: 'Uniquement femmes',
    };

    const getMappedValues = (arr: string[], mapping: Record<string, string>) =>
      arr
        .map((curr) => mapping[curr])
        .filter(Boolean)
        .join(';');

    const ageString = getMappedValues(age, ageMapping);
    const handicapsString = getMappedValues(handicaps, handicapsMapping);
    const languagesString = getMappedValues(languages, languageMapping);
    const otherString = getMappedValues(other, otherMapping);

    return [ageString, handicapsString, languagesString, otherString].filter(Boolean).join(';');
  }

  private checkAccessConditions(accessModality: string[]): string {
    if (!accessModality) {
      return '';
    }

    if (accessModality.includes('accesLibre')) {
      return 'Gratuit : Je peux accéder gratuitement au lieu et à ses services';
    } else {
      return '';
    }
  }

  private formatServices(structCat: StructureCategories): string {
    if (!structCat) {
      return '';
    }

    let result = '';

    for (const [stringToAppend, servicesToMatch] of Object.entries(config)) {
      for (const categoryName of Object.keys(structCat)) {
        if (servicesToMatch.categories.includes(categoryName)) {
          const matchFound = servicesToMatch.values.some((service) => structCat[categoryName].includes(service));
          if (matchFound) {
            result += stringToAppend + ';';
          }
        }
      }
    }

    return result.slice(0, -1);
  }

  private formatDate(date: Date | undefined | null): string {
    if (!date) {
      return '1970-01-01';
    }
    return date.toISOString().substring(0, 10);
  }

  private formatAddress(address: { numero?: string; street: string }): string {
    if (address.numero) {
      return address.numero + ' ' + address.street;
    } else {
      return address.street;
    }
  }

  private convertToOSMFormat(openingHours: Record<string, any>): string {
    const openingHoursDays = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'];
    const daysOfWeek = ['Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa', 'Su'];
    const timetable = [];

    // iterate over each day of the week
    for (let i = 0; i < daysOfWeek.length; i++) {
      const dayOfWeekSource = openingHoursDays[i];
      const dayOfWeekTarget = daysOfWeek[i];
      const dayHours = openingHours[dayOfWeekSource];
      // if the day is open, convert its hours to a string format
      if (dayHours.open) {
        const dayTimetable = {
          day: dayOfWeekTarget,
          osmHours: dayHours.time.map((time) => `${time.opening}-${time.closing}`).join(','),
        };
        timetable.push(dayTimetable);
      }
    }
    const osmOpeningHours: string = toOsmOpeningHours(timetable);
    return osmOpeningHours;
  }
}
