import { getModelToken } from '@nestjs/mongoose';
import { Test } from '@nestjs/testing';
import { of } from 'rxjs';
import { DataGouvStructure } from '../interfaces/data-gouv-structure.interface';
import { HttpService } from '@nestjs/axios';
import { AxiosResponse, InternalAxiosRequestConfig } from 'axios';
import { mockGouvStructure, mockGouvStructureToResinFormat } from '../../../test/mock/data/gouvStructures.mock.data';
import { StructuresSearchService } from './structures-search.service';
import { StructuresService } from './structures.service';
import { StructuresImportService } from './structures-import.service';
import { MailerService } from '../../mailer/mailer.service';

describe('StructuresImportService', () => {
  let structuresImportService: StructuresImportService;
  let structuresSearchService: StructuresSearchService;
  let structuresService: StructuresService;

  const httpServiceMock = {
    get: jest.fn(),
  };

  beforeEach(async () => {
    const module = await Test.createTestingModule({
      providers: [
        StructuresImportService,
        {
          provide: HttpService,
          useValue: {
            get: jest.fn(),
          },
        },
        {
          provide: StructuresSearchService,
          useValue: {
            indexStructure: jest.fn().mockResolvedValue(undefined),
          },
        },
        {
          provide: StructuresService,
          useValue: {
            updateDenormalizedFields: jest.fn().mockResolvedValue(undefined),
            sendAdminNotificationAfterImportingStructure: jest.fn().mockImplementation(() => of(undefined)),
          },
        },
        {
          provide: MailerService,
          useValue: {
            config: {
              templates: {
                adminStructureImport: { ejs: null, json: null },
              },
            },
          },
        },
        {
          provide: getModelToken('Structure'),
          useValue: {
            countDocuments: jest.fn(),
            new: jest.fn(),
            save: jest.fn(),
          },
        },
        {
          provide: HttpService,
          useValue: httpServiceMock,
        },
      ],
    }).compile();

    structuresImportService = module.get<StructuresImportService>(StructuresImportService);
    structuresSearchService = module.get<StructuresSearchService>(StructuresSearchService);
    structuresService = module.get<StructuresService>(StructuresService);
  });

  describe('importDataGouvStructures', () => {
    it('should create a new structure when doesAlreadyExist returns false', async () => {
      const mockedAxiosResponse: AxiosResponse<DataGouvStructure[]> = {
        data: [mockGouvStructure],
        status: 200,
        statusText: 'OK',
        headers: {},
        config: {} as InternalAxiosRequestConfig,
      };

      const saveSpy = jest
        .fn()
        .mockResolvedValue({ id: 'b145231d-7388-4444-a54e-138298a4fff9|res-in-606715119f33ab0013a1cbad' });
      const mockStructureModel = jest.fn().mockImplementation(() => ({ save: saveSpy }));
      (structuresImportService as any).structureModel = mockStructureModel;

      const formatToResinSchemaSpy = jest
        .spyOn(structuresImportService as any, 'formatToResinSchema')
        .mockResolvedValueOnce('called');
      const doesAlreadyExistSpy = jest
        .spyOn(structuresImportService as any, 'doesAlreadyExist')
        .mockResolvedValueOnce(false);
      httpServiceMock.get.mockImplementationOnce(() => of(mockedAxiosResponse));

      jest.spyOn(structuresSearchService, 'indexStructure').mockResolvedValueOnce(undefined);

      jest.spyOn(structuresService, 'updateDenormalizedFields').mockResolvedValueOnce(undefined);

      await structuresImportService.importDataGouvStructures();

      expect(httpServiceMock.get).toHaveBeenCalled();
      expect(formatToResinSchemaSpy).toHaveBeenCalled();
      expect(doesAlreadyExistSpy).toHaveBeenCalled();
      expect(saveSpy).toHaveBeenCalled();
    });
  });

  describe('formatToResinSchema', () => {
    it('should transform data.gouv structures to resin format', async () => {
      const result = await structuresImportService['formatToResinSchema'](mockGouvStructure);
      expect(result).toEqual(mockGouvStructureToResinFormat);
    });
  });

  describe('convertFromOSMFormat', () => {
    it('should convert simple weekday format', () => {
      const osmFormat = 'Mo-Fr 08:30-12:30';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.monday.open).toBe(true);
      expect(result.monday.time).toEqual([{ opening: '08:30', closing: '12:30' }]);
      expect(result.tuesday.time).toEqual([{ opening: '08:30', closing: '12:30' }]);
      expect(result.wednesday.time).toEqual([{ opening: '08:30', closing: '12:30' }]);
      expect(result.thursday.time).toEqual([{ opening: '08:30', closing: '12:30' }]);
      expect(result.friday.time).toEqual([{ opening: '08:30', closing: '12:30' }]);
      expect(result.saturday.open).toBe(false);
      expect(result.sunday.open).toBe(false);
    });

    it('should handle a day marked as off', () => {
      const osmFormat = 'Mo-Fr 08:30-12:30; We off';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.wednesday.open).toBe(false);
    });

    it('should handle days marked as off', () => {
      const osmFormat = 'Mo-Fr 08:30-12:30; We,Th off';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.wednesday.open).toBe(false);
      expect(result.thursday.open).toBe(false);
    });

    it('should handle multiple time ranges', () => {
      const osmFormat = 'Mo-Fr 08:30-12:30,13:30-17:30';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.monday.time).toEqual([
        { opening: '08:30', closing: '12:30' },
        { opening: '13:30', closing: '17:30' },
      ]);
    });

    it('should handle individual days with multiple time ranges', () => {
      const osmFormat = 'Mo,We 08:30-12:30,13:30-17:30';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.monday.time).toEqual([
        { opening: '08:30', closing: '12:30' },
        { opening: '13:30', closing: '17:30' },
      ]);
      expect(result.wednesday.time).toEqual([
        { opening: '08:30', closing: '12:30' },
        { opening: '13:30', closing: '17:30' },
      ]);
    });

    it('should handle complex OSM format', () => {
      const osmFormat = 'Mo-Fr 08:30-12:30,13:30-17:30; We,Th off';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.monday.time).toEqual([
        { opening: '08:30', closing: '12:30' },
        { opening: '13:30', closing: '17:30' },
      ]);
      expect(result.wednesday.open).toBe(false);
      expect(result.thursday.open).toBe(false);
    });

    it('should handle a very complex OSM format', () => {
      const osmFormat = 'Mo-Sa 08:30-11:45; Fr 08:30-11:45,14:30-16:45; We off ';
      const result = structuresImportService['convertFromOSMFormat'](osmFormat);

      expect(result.monday.time).toEqual([{ opening: '08:30', closing: '11:45' }]);
      expect(result.wednesday.open).toBe(false);
      expect(result.friday.time).toEqual([
        { opening: '08:30', closing: '11:45' },
        { opening: '14:30', closing: '16:45' },
      ]);
    });
  });
});
