import { Logger } from '@nestjs/common';
import { ConfigModule } from '@nestjs/config';
import { Test } from '@nestjs/testing';
import { StructuresForSearchServiceMock } from '../../../test/mock/services/structures-for-search.mock.service';
import { MailerModule } from '../../mailer/mailer.module';
import { SearchModule } from '../../search/search.module';
import { StructuresSearchService } from './structures-search.service';
import { StructuresService } from './structures.service';

describe('StructuresSearchService', () => {
  let structureSearchService: StructuresSearchService;

  beforeEach(async () => {
    const module = await Test.createTestingModule({
      imports: [MailerModule, SearchModule, ConfigModule],
      providers: [
        StructuresSearchService,
        {
          provide: StructuresService,
          useClass: StructuresForSearchServiceMock,
        },
      ],
    }).compile();

    structureSearchService = module.get<StructuresSearchService>(StructuresSearchService);
    structureSearchService['index'] = 'structures-unit-test';
    await structureSearchService.dropIndex();
    await structureSearchService.createStructureIndex();

    // init str
    const structuresForSearchService = new StructuresForSearchServiceMock();
    const structures = structuresForSearchService.findAll();
    await Promise.all(
      structures.map((structure) => {
        structureSearchService.indexStructure(structure);
      })
    );

    // wait for the new structures to be indexed before search
    await structureSearchService.refreshIndexStructure();
    await new Promise((r) => setTimeout(r, 2000));
  }, 10000);

  it('should be defined', () => {
    expect(structureSearchService).toBeDefined();
  });

  describe('search method', () => {
    it('should find maisons de la métropole', async () => {
      const res = await structureSearchService.search('maison de la');
      expect(res[0].structureName).toContain('Maison de la Métropole');
      expect(res[1].structureName).toContain('Maison de la Métropole');
    });

    it('should find metropole', async () => {
      const res = await structureSearchService.search('metropole');
      expect(res[0].structureName).toContain('Métropole');
    });

    it('should find structure with stopword "son"', async () => {
      const res = await structureSearchService.search('son');
      expect(res[0].structureName).toContain('Le Son du Clic');
    });

    it('should find structure with stopword "mais"', async () => {
      const res = await structureSearchService.search('mais');
      expect(res[0].structureName.toLowerCase()).toContain('maison');
    });

    it('should find structure with stopword "le"', async () => {
      const res = await structureSearchService.search('le');
      expect(res.length).toBeGreaterThanOrEqual(1);
    });

    it('should find structure with partial word "oull"', async () => {
      const res = await structureSearchService.search('oull');
      expect(res[0].structureName.toLowerCase()).toContain('oullins');
    });

    it('should find structure with partial word "vill"', async () => {
      const res = await structureSearchService.search('vill');
      expect(res[0].structureName.toLowerCase()).toContain('villeurbanne');
    });

    it('should find structure with partial word "asso"', async () => {
      const res = await structureSearchService.search('asso');
      Logger.log(JSON.stringify(res));
      expect(res[0].structureName.toLowerCase()).toContain('association');
    });

    it('should find text in description', async () => {
      const res = await structureSearchService.search('liseuse');
      expect(res.length).toBe(1);
      expect(res[0].structureName).toContain("Médiathèque d'Ecully");
    });

    it('should not find text in description when search by name', async () => {
      const res = await structureSearchService.search('liseuse', ['structureName']);
      expect(res.length).toBe(0);
    });

    it('should find structure when search by name', async () => {
      const res = await structureSearchService.search('le son', ['structureName']);
      expect(res[0].structureName).toContain('Le Son du Clic');
    });
  });

  it('should drop index', async () => {
    const res = await structureSearchService.dropIndex();
    expect(res).toBeTruthy();
  });
});
