import {
  IndicesCreateResponse,
  IndicesRefreshResponse,
  IndicesResponseBase,
  UpdateResponse,
} from '@elastic/elasticsearch/lib/api/types';
import { Injectable, Logger } from '@nestjs/common';
import { ElasticsearchService } from '@nestjs/elasticsearch';
import { escapeElasticsearchQuery, es_settings_homemade_french } from '../../shared/utils';
import { StructureSearchBody } from '../interfaces/structure-search-body.interface';
import { StructureDocument } from '../schemas/structure.schema';

@Injectable()
export class StructuresSearchService {
  private index = 'structures';
  private readonly logger = new Logger(StructuresSearchService.name);

  constructor(private readonly elasticsearchService: ElasticsearchService) {}

  public async indexStructure(structure: StructureDocument): Promise<StructureDocument> {
    this.logger.debug(`indexStructure`);
    await this.elasticsearchService.index<StructureSearchBody>({
      index: this.index,
      id: structure._id,
      document: this.formatIndexBody(structure),
    });
    return structure;
  }

  private formatIndexBody(structure: StructureDocument): StructureSearchBody {
    return {
      structureName: structure.structureName,
      structureTypeName: structure.structureType?.name,
      structureTypeValue: structure.structureType?.value,
      structureId: structure._id,
      address: structure.address,
      description: structure.description,
    };
  }

  public async createStructureIndex(): Promise<IndicesCreateResponse> {
    this.logger.debug(`createStructureIndex`);
    // use custom analyzer with minimal_french stemmer to avoid default light_french stemmer problems (oullins -> oulin, "oull" not found)
    // don't use stopwords (le, la, de, son, etc.) for structureName (to find "maison de la", "le son du clic", etc.)
    return this.elasticsearchService.indices.create({
      index: this.index,
      mappings: {
        dynamic_templates: [
          {
            strings: {
              match_mapping_type: 'string',
              mapping: {
                type: 'text',
                analyzer: 'homemade_french',
                search_analyzer: 'homemade_french',
              },
            },
          },
        ],
        properties: {
          structureName: {
            type: 'text',
            analyzer: 'homemade_french_stopless',
            search_analyzer: 'homemade_french_stopless',
          },
        },
      },
      settings: es_settings_homemade_french,
    });
  }

  public async dropIndex(): Promise<IndicesResponseBase> {
    this.logger.debug(`dropIndex`);
    if (
      await this.elasticsearchService.indices.exists({
        index: this.index,
      })
    ) {
      return await this.elasticsearchService.indices.delete({
        index: this.index,
      });
    }
  }

  public async deleteIndexStructure(structure: StructureDocument): Promise<StructureDocument> {
    this.logger.debug(`deleteIndexStructure`);
    this.elasticsearchService.delete({
      index: this.index,
      id: structure._id,
    });
    return structure;
  }

  public async refreshIndexStructure(): Promise<IndicesRefreshResponse> {
    this.logger.debug(`refreshIndexStructure`);
    return this.elasticsearchService.indices.refresh({
      index: this.index,
    });
  }

  public async search(searchString: string, fields?: string[]): Promise<StructureSearchBody[]> {
    this.logger.debug(`search ${searchString} | fields : ${fields}`);
    searchString = searchString ? searchString + '*' : '*';
    const { hits } = await this.elasticsearchService.search<StructureSearchBody>({
      index: this.index,
      body: {
        from: 0,
        size: 5000,
        query: {
          query_string: {
            analyze_wildcard: true,
            query: escapeElasticsearchQuery(searchString),
            fields: fields
              ? fields
              : [
                  'structureName^5',
                  'structureTypeName^5',
                  'structureTypeValue^5',
                  'address.commune^10',
                  'address.postcode^5',
                  'description',
                ],
            fuzziness: 'AUTO',
          },
        },
      },
    });
    const sortedHits = hits.hits.filter(function (elem) {
      return elem._score >= hits.max_score / 1.5;
    });
    return sortedHits.map((item) => item._source);
  }

  public async update(structure: StructureDocument, id: string): Promise<UpdateResponse<unknown>> {
    this.logger.debug(`Update ES structure: ${structure.structureName}`);
    return this.elasticsearchService.update({
      index: this.index,
      id: id,
      doc: this.formatIndexBody(structure),
    });
  }
}
