import { HttpService } from '@nestjs/axios';
import { HttpException, HttpStatus, Inject, Injectable, Logger, forwardRef } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import { Cron, CronExpression } from '@nestjs/schedule';
import { AxiosResponse } from 'axios';
import * as ejs from 'ejs';
import * as _ from 'lodash';
import { DateTime } from 'luxon';
import { FilterQuery, Model, Types } from 'mongoose';
import { lastValueFrom } from 'rxjs';
import { PendingStructureDto } from '../../admin/dto/pending-structure.dto';
import { Categories } from '../../categories/schemas/categories.schema';
import { Module } from '../../categories/schemas/module.class';
import { CategoriesService } from '../../categories/services/categories.service';
import { MailerService } from '../../mailer/mailer.service';
import { PersonalOfferDocument } from '../../personal-offers/schemas/personal-offer.schema';
import { sanitize, hasAdminRole } from '../../shared/utils';
import { IUser } from '../../users/interfaces/user.interface';
import { User } from '../../users/schemas/user.schema';
import { JobsService } from '../../users/services/jobs.service';
import { UsersService } from '../../users/services/users.service';
import { PhotonAddress } from '../common/photonAddress.type';
import { PhotonResponse } from '../common/photonResponse.type';
import { depRegex } from '../common/regex';
import { StructureDto } from '../dto/structure.dto';
import { UpdateStructureDto } from '../dto/update-structure.dto';
import { EquipmentsServicesEnum } from '../enum/equipmentsServices';
import { PhotonPoints } from '../interfaces/photon-response.interface';
import { Address } from '../schemas/address.schema';
import { Structure, StructureDocument } from '../schemas/structure.schema';
import { StructuresSearchService } from './structures-search.service';

@Injectable()
export class StructuresService {
  private readonly logger = new Logger(StructuresService.name);

  constructor(
    private readonly httpService: HttpService,
    @Inject(forwardRef(() => UsersService))
    private readonly userService: UsersService,
    private readonly mailerService: MailerService,
    private structuresSearchService: StructuresSearchService,
    private categoriesService: CategoriesService,
    private jobsService: JobsService,
    @InjectModel(Structure.name) private structureModel: Model<StructureDocument>
  ) {}

  async initiateStructureIndex(): Promise<StructureDocument[]> {
    this.logger.log('Reset structures indexes');
    await this.structuresSearchService.dropIndex();
    await this.structuresSearchService.createStructureIndex();
    return this.populateES();
  }

  async searchForStructures(
    text: string,
    filters?: Array<any>,
    fields?: string[],
    onlyOffersWithAppointment?: boolean,
    limit?: number,
    inseeCodes?: string[]
  ): Promise<StructureDocument[]> {
    this.logger.debug(
      `searchForStructures : ${text} | filters: ${JSON.stringify(filters)} | fields : ${JSON.stringify(fields)} | `
    );
    const results = await this.structuresSearchService.search(text, fields);
    const ids = results.map((result) => result.structureId);
    let structures: StructureDocument[];
    const andFilters = filters ? filters[0] : null;
    const orFilters = filters ? filters[1] : null;

    if (!ids.length) {
      return [];
    }

    // Separate filters on offers which will be filtered using personalOffers below
    const [andFiltersOnOffers, andFiltersNotOnOffers] = _.partition(andFilters, function (filter) {
      return 'onlineProcedures' in filter || 'baseSkills' in filter || 'advancedSkills' in filter;
    });

    // Construct a global $and condition that can encompass multiple $and and $or clauses
    let queryConditions: FilterQuery<StructureDocument>[] = [
      { deletedAt: { $exists: false } },
      { accountVerified: true },
    ];
    if (andFiltersNotOnOffers?.length) {
      queryConditions.push({
        $and: [...this.parseFilter(andFiltersNotOnOffers)],
      });
    }
    if (orFilters?.length) {
      queryConditions = this.addOrConditions(orFilters, queryConditions);
    }
    if (inseeCodes?.length > 0) {
      queryConditions.push({
        $or: [{ 'address.inseeCode': { $in: inseeCodes } }],
      });
    }

    // Match ids from Elasticsearch text search with mongoDB query with filters
    structures = await this.structureModel
      .find({
        _id: { $in: ids },
        $and: queryConditions,
      })
      .populate({ path: 'personalOffers', model: 'PersonalOffer' })
      .populate('structureType')
      .limit(limit)
      .exec();

    // Filter offers using structure offers and structure members personalOffers
    if (andFiltersOnOffers.length) {
      structures = await this.filterOnOffers(structures, andFiltersOnOffers, onlyOffersWithAppointment);
    }

    // Sort the result according to ElasticSearch order
    return structures.sort((a, b) => ids.indexOf(a.id) - ids.indexOf(b.id));
  }

  /**
   * For "or" filters of different types, add to queryConditions a different "or" filter for each type
   * (For example for "or" filters "ctm" and "freeWorkShop" : cf. issue #620 )
   */
  private addOrConditions(
    orFilters,
    queryConditions: FilterQuery<StructureDocument>[]
  ): FilterQuery<StructureDocument>[] {
    // Object to hold arrays for each key (type of filter)
    const grouped = {};

    orFilters.forEach((item) => {
      // Get the key (assumes each item has only one key)
      const key = Object.keys(item)[0];
      const value = item[key];

      // Initialize the array for the key if it doesn't exist
      if (!grouped[key]) {
        grouped[key] = [];
      }

      // Push the key-value pair into the array
      grouped[key].push({ [key]: value });
    });

    // Create $or conditions for each group (type of filter)
    for (const key in grouped) {
      queryConditions.push({
        $or: this.parseFilter(grouped[key]),
      });
    }

    return queryConditions;
  }

  /**
   * set structure offers and structure social workers personalOffers in non-persistant property structure.categoriesWithPersonalOffers
   */
  public async setCategoriesWithPersonalOffers(structure: StructureDocument, onlyOffersWithAppointment: boolean) {
    if (onlyOffersWithAppointment) {
      // Ignore structure offers (for orientation rdv, structure offers must be ignored: we filter only on personalOffers of the structure)
      structure.categoriesWithPersonalOffers = {
        onlineProcedures: [],
        baseSkills: [],
        advancedSkills: [],
      };
    } else {
      // Get structure offers
      structure.categoriesWithPersonalOffers = {
        onlineProcedures: structure.categories?.onlineProcedures || [],
        baseSkills: structure.categories?.baseSkills || [],
        advancedSkills: structure.categories?.advancedSkills || [],
      };
    }

    // Promise.all to wait for execution of all asynchronous operations
    await Promise.all(
      (structure.personalOffers || []).map(async (personalOffer) => {
        if (!personalOffer.categories) {
          Logger.log(`personalOffer not populated for structure ${structure.structureName} : ${personalOffer}`);
          return;
        }

        // If we only want personalOffers from user with appointment
        if (onlyOffersWithAppointment) {
          const user = await this.userService.findByPersonalOfferId(personalOffer._id);
          if (!user) {
            Logger.log(
              `No user with personalOffer ${personalOffer._id} for structure ${structure.structureName} (${structure._id}) !`,
              StructuresService.name
            );
            return;
          }
          if (!user.structuresLink.includes(structure._id)) {
            Logger.log(
              `Inconsistent personalOffer skipped: user ${user.name} ${user.surname} (${user._id}) has personalOffer ${personalOffer._id} for structure ${structure.structureName} (${structure._id}) but is not (anymore) member of the structure !`,
              StructuresService.name
            );
            return;
          }
          // If user is without appointment, skip this personalOffer
          if (!User.getWithAppointment(user)) {
            return;
          }
        }

        // use lodash _.union fonction to concat array without duplicates
        structure.categoriesWithPersonalOffers.onlineProcedures = _.union(
          structure.categoriesWithPersonalOffers.onlineProcedures,
          personalOffer.categories.onlineProcedures
        );
        structure.categoriesWithPersonalOffers.baseSkills = _.union(
          structure.categoriesWithPersonalOffers.baseSkills,
          personalOffer.categories.baseSkills
        );
        structure.categoriesWithPersonalOffers.advancedSkills = _.union(
          structure.categoriesWithPersonalOffers.advancedSkills,
          personalOffer.categories.advancedSkills
        );
      })
    );
  }

  public async create(email: string, structure: StructureDto): Promise<Structure> {
    const user = await this.userService.findOne(email);
    if (!user) {
      throw new HttpException('Invalid profile', HttpStatus.NOT_FOUND);
    }
    // Handle equipments
    // Value is forced to true because there is no verification needed now. Cannot be removed because of old data prod
    structure.accountVerified = true;
    const createdStructure = new this.structureModel(structure);
    createdStructure._id = new Types.ObjectId();
    createdStructure.structureName = createdStructure.structureName.trim();
    createdStructure.permalink = await this.getAvailablePermalink(createdStructure.structureName);
    createdStructure.categories.selfServiceMaterial = this.getSelfServiceMaterial(createdStructure);
    await createdStructure.save();
    await this.setStructurePosition(createdStructure).then(async (structureWithPosition: StructureDocument) => {
      return this.structuresSearchService.indexStructure(
        await this.structureModel
          .findByIdAndUpdate(new Types.ObjectId(createdStructure._id), structureWithPosition, { new: true })
          .populate({ path: 'personalOffers', model: 'PersonalOffer' })
          .populate('structureType')
          .exec()
      );
    });
    user.structuresLink.push(createdStructure._id);
    user.unattachedSince = null;
    await user.save();

    // Update denormalized fields and CNFS label of the structure
    const updatedStructure = await this.updateDenormalizedFields(new Types.ObjectId(createdStructure._id));

    // Send admin notification mail
    this.sendAdminStructureNotification(
      createdStructure,
      this.mailerService.config.templates.adminStructureCreate.ejs,
      this.mailerService.config.templates.adminStructureCreate.json
    );

    return updatedStructure;
  }

  public async getAvailablePermalink(structureName: string): Promise<string> {
    let permalink = sanitize(structureName);

    // If already exists, add a number to the end of the permalink
    const exists = await this.findByPermalink(permalink);
    if (exists) {
      let count = 1;
      while (await this.findByPermalink(`${permalink}-${count}`)) {
        count++;
      }
      permalink = `${permalink}-${count}`;
    }

    return permalink;
  }

  public async search(searchString: string, filters?: Array<any>): Promise<Structure[]> {
    if (searchString && filters) {
      return this.structureModel
        .find({
          $and: [
            ...this.parseFilter(filters),
            { $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true },
          ],
        })
        .populate({ path: 'personalOffers', model: 'PersonalOffer' })
        .populate('structureType')
        .exec();
    } else if (filters) {
      return this.structureModel
        .find({ $and: [{ $and: this.parseFilter(filters), deletedAt: { $exists: false }, accountVerified: true }] })
        .populate({ path: 'personalOffers', model: 'PersonalOffer' })
        .populate('structureType')
        .exec();
    } else {
      return this.structureModel
        .find({
          $and: [{ $or: [{ $text: { $search: searchString }, deletedAt: { $exists: false }, accountVerified: true }] }],
        })
        .populate({ path: 'personalOffers', model: 'PersonalOffer' })
        .populate('structureType')
        .exec();
    }
  }

  /**
   * Parse equipments. Ex : if structure doesn't have computers but nbComputer changes, then the key is added to equipmentsAndServices
   */
  public getSelfServiceMaterial(structure: StructureDocument): string[] {
    Logger.debug(`getSelfServiceMaterial of ${structure.structureName}`, StructuresService.name);
    if (!structure.categories.selfServiceMaterial) {
      structure.categories.selfServiceMaterial = [];
    }
    if (structure.nbComputers > 0) {
      structure.categories.selfServiceMaterial.push(EquipmentsServicesEnum.computer);
    } else {
      const index = structure.categories.selfServiceMaterial.indexOf(EquipmentsServicesEnum.computer);
      if (index !== -1) structure.categories.selfServiceMaterial.splice(index, 1);
    }
    if (structure.nbPrinters > 0) {
      structure.categories.selfServiceMaterial.push(EquipmentsServicesEnum.printer);
    } else {
      const index = structure.categories.selfServiceMaterial.indexOf(EquipmentsServicesEnum.printer);
      if (index !== -1) structure.categories.selfServiceMaterial.splice(index, 1);
    }
    if (structure.nbScanners > 0) {
      structure.categories.selfServiceMaterial.push(EquipmentsServicesEnum.scanner);
    } else {
      const index = structure.categories.selfServiceMaterial.indexOf(EquipmentsServicesEnum.scanner);
      if (index !== -1) structure.categories.selfServiceMaterial.splice(index, 1);
    }

    // Remove duplicate if exists
    return [...new Set(structure.categories.selfServiceMaterial)];
  }

  /**
   * @param deleted boolean to get deleted structures
   */
  public async findAll(deleted = false): Promise<StructureDocument[]> {
    this.logger.debug('findAll');
    return this.structureModel
      .find({ deletedAt: { $exists: deleted }, accountVerified: true })
      .populate('structureType')
      .exec();
  }

  public async findAllUnclaimed(): Promise<StructureDocument[]> {
    return await this.structureModel
      .find({ deletedAt: { $exists: false }, hasNoUserDN: true })
      .populate({ path: 'personalOffers', model: 'PersonalOffer' })
      .populate('structureType')
      .exec();
  }

  public async populateES(): Promise<StructureDocument[]> {
    const structures = await this.structureModel
      .find({ deletedAt: { $exists: false } })
      .populate({ path: 'personalOffers', model: 'PersonalOffer' })
      .populate('structureType')
      .exec();
    await Promise.all(
      structures.map((structure: StructureDocument) => {
        this.structuresSearchService.indexStructure(structure);
      })
    );
    return structures;
  }

  /**
   * Find all structure only returning structures who consented to data sharing
   */
  public async findConsentingStructures(categories: Categories[]): Promise<StructureDocument[]> {
    this.logger.debug('Find all structures, only returning structures who consented to data sharing');
    return (
      await this.structureModel
        .find({ deletedAt: { $exists: false }, accountVerified: true, $where: 'this.dataShareConsentDate != null' })
        .select('-_id -accountVerified -otherDescription -dataShareConsentDate')
        .exec()
    ).map((structure) => {
      const repositoryKeys = categories.map((category) => category.id);
      repositoryKeys.forEach((el) => {
        // Add referentiel
        if (structure.categories[el] !== undefined) {
          structure.categories[el] = this.mapModules(
            structure.categories[el],
            categories.find((category) => category.id === el).modules
          );
        }
      });
      return structure;
    });
  }

  /**
   * Map modules labels to ids
   */
  public mapModules(structureModule: string[], baseModule: Module[]): string[] {
    if (structureModule.length === 0) {
      return [];
    }
    return structureModule.map((id) => {
      const formattedText = baseModule.find((module) => {
        return module.id === id;
      })?.name;
      return formattedText || id;
    });
  }

  public async update(idStructure: string, updatedFields: UpdateStructureDto, user: User): Promise<Structure> {
    this.logger.debug(`Updating structure ${idStructure}`);
    const oldStructure = await this.findOne(idStructure);

    // Update position only if address has changed (to not erase coord which may have been set manually in database)
    if (updatedFields.address && !Address.equals(updatedFields.address, oldStructure.address)) {
      oldStructure.address = updatedFields.address;
      await this.setStructurePosition(oldStructure);
    }

    // Store updated categories because it will be erased by Object.assign which is a shallow copy
    const oldCategories = { ...oldStructure.categories };
    const updatedCategories = { ...updatedFields.categories };
    const oldTimestamp = oldStructure.updatedAt;
    const deepClone: StructureDocument = Object.assign(oldStructure, updatedFields);

    // Update structure categories in order to not override it
    if (updatedCategories) {
      deepClone.categories = Object.assign(deepClone.categories, oldCategories);
      Object.keys(updatedCategories).forEach((key) => {
        deepClone.categories[key] = updatedCategories[key];
      });
    }

    // Keep previous timestamp (because shallow copy updates timestamps of both objects) only if the user is an admin so as not to disrupt the mechanism to get the user to update their structures
    if (hasAdminRole(user)) {
      deepClone.updatedAt = oldTimestamp;
    }

    if (updatedFields.structureName) {
      deepClone.structureName = deepClone.structureName.trim();
    }
    const parsedStructure = deepClone;
    parsedStructure.categories.selfServiceMaterial = this.getSelfServiceMaterial(deepClone);

    const result = await this.structureModel
      .findByIdAndUpdate(new Types.ObjectId(idStructure), parsedStructure, {
        new: true,
        timestamps: !hasAdminRole(user),
      })
      .exec();

    if (!result) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }

    const responseStructure = await this.findOne(idStructure);
    await this.structuresSearchService.update(responseStructure, idStructure);
    if (!hasAdminRole(user)) {
      this.sendAdminStructureNotification(
        result,
        this.mailerService.config.templates.structureModificationNotification.ejs,
        this.mailerService.config.templates.structureModificationNotification.json
      );
      this.userService.removeOutdatedStructureFromArray(idStructure);
    }

    return responseStructure;
  }

  public async findOne(idParam: string): Promise<StructureDocument> {
    this.logger.debug(`findOne: ${idParam}`);
    const structureQuery = this.structureModel.findById(new Types.ObjectId(idParam));
    return this.findQueryExec(structureQuery);
  }

  public async findByPermalink(permalink: string): Promise<StructureDocument> {
    this.logger.debug(`findByPermalink: ${permalink}`);
    const structureQuery = this.structureModel.findOne({ permalink });
    return this.findQueryExec(structureQuery);
  }

  public async getInseeCode(postcode: string, municipality: string): Promise<string | null> {
    const req = `https://apicarto.ign.fr/api/codes-postaux/communes/${postcode}`;
    this.logger.debug(`Request : ${req}`);

    let res;
    try {
      res = await lastValueFrom(this.httpService.get(encodeURI(req)));
    } catch (e) {
      this.logger.error(`Request ${encodeURI(req)} failed: ${e}`);
      return null;
    }

    if (res.data?.length) {
      if (res.data.length == 1) {
        this.logger.debug(`Insee code ${res.data[0].codeCommune} for ${municipality}`);
        return res.data[0].codeCommune;
      } else {
        // if several municipality for this postcode, find for which municipality we are looking for the insee code
        const communes = res.data.filter((commune) => commune.nomCommune == municipality);
        if (communes.length) {
          this.logger.debug(`Insee code ${communes[0].codeCommune} for ${municipality}`);
          return communes[0].codeCommune;
        } else {
          this.logger.error(
            `Insee code of commune ${municipality} not found for postcode ${postcode} in ${JSON.stringify(
              res.data
            )} (${req})`
          );
          return null;
        }
      }
    } else {
      return null;
    }
  }

  /**
   * Get city formatted without "Arrondissement"
   */
  public getFormattedCity(city: string, postcode: string) {
    let formattedCity: string = city;
    if (city?.startsWith('Lyon')) {
      const arrondissement = parseInt(postcode.slice(-2));
      if (arrondissement == 1) {
        formattedCity = `Lyon 1er`;
      } else {
        formattedCity = `Lyon ${arrondissement}e`;
      }
    }
    return formattedCity;
  }

  /**
   * Search structure address based on data search WS
   */
  public async searchAddress(data: { searchQuery: string }): Promise<{ features: PhotonPoints[] }> {
    const reqBan = `https://download.data.grandlyon.com/geocoding/photon/api?q=${data.searchQuery}&lang=fr&limit=500&osm_tag=:!construction&osm_tag=:!bus_stop`;
    const reqBal = `https://download.data.grandlyon.com/geocoding/photon-bal/api?q=${data.searchQuery}&lang=fr&limit=500&osm_tag=:!construction&osm_tag=:!bus_stop`;

    const requestGroup = async (url: string): Promise<AxiosResponse<{ features: PhotonPoints[] }>> => {
      const response$ = this.httpService.get<{ features: PhotonPoints[] }>(encodeURI(url), {
        headers: { 'Content-Type': 'application/json' },
      });

      const response = await lastValueFrom(response$);
      response.data.features = response.data.features
        .filter((doc) => doc.properties.postcode && doc.properties.postcode.match(depRegex))
        .sort((a, b) => (b.properties.housenumber ? 1 : -1));

      return response;
    };

    const [reqBanRes, reqBalRes] = await Promise.all([requestGroup(reqBan), requestGroup(reqBal)]).catch((e) => {
      this.logger.error('searchAddress photon requests failed');
      this.logger.error(e);
      throw new HttpException(`Address service temporary not available : ${e.message}`, HttpStatus.SERVICE_UNAVAILABLE);
    });
    const mergedArray = [...reqBalRes.data.features, ...reqBanRes.data.features];

    const duplicateFreeArray = _.unionWith(mergedArray, (a, b) => {
      /**
       * Remove accents from a string by decomposing accented characters and removing diacritical marks.
       * @param {string} str - The input string.
       * @returns {string} The string with accents removed.
       */
      const removeAccents = (str: string) => {
        if (!str) return '';
        // Decompose accented characters into base characters and diacritical marks, then remove the diacritical marks using a regex pattern.
        return str.normalize('NFD').replace(/[\u0300-\u036f]/g, '');
      };
      return (
        a.properties.postcode === b.properties.postcode &&
        ((a.properties.name &&
          (removeAccents(a.properties.name) === removeAccents(b.properties.name) ||
            removeAccents(a.properties.name) === removeAccents(b.properties.street))) ||
          (a.properties.street &&
            (removeAccents(a.properties.street) === removeAccents(b.properties.name) ||
              removeAccents(a.properties.street) === removeAccents(b.properties.street)))) &&
        (a.properties.housenumber === b.properties.housenumber ||
          (!a.properties.housenumber && !b.properties.housenumber))
      );
    });

    duplicateFreeArray.forEach((feature) => {
      feature.properties.city = this.getFormattedCity(feature.properties.city, feature.properties.postcode);
    });

    return { features: duplicateFreeArray };
  }

  /**
   * Count every value occurrence of a given key
   * @param key structure key
   */
  public async countByStructureKey(key: string, selected: { id: string; text: string }[]) {
    const uniqueElements = await this.structureModel.distinct(key).exec();
    return Promise.all(
      uniqueElements.map(async (value) => {
        const keyList: FilterQuery<StructureDocument>[] = [];
        keyList.push({
          [key]: { $elemMatch: { $eq: value } },
          deletedAt: { $exists: false },
          accountVerified: true,
        });
        if (selected && selected.length > 0) {
          for (const val of selected) {
            keyList.push({
              [val.text]: { $elemMatch: { $eq: val.id } },
              deletedAt: { $exists: false },
              accountVerified: true,
            });
          }
        }
        return {
          id: value,
          count: await this.structureModel
            .countDocuments({
              $and: keyList,
            })
            .exec(),
        };
      })
    );
  }

  public async getCoord(
    numero: string,
    address: string,
    commune: string,
    scope: string
  ): Promise<AxiosResponse<PhotonResponse>> {
    const req =
      `https://download.data.grandlyon.com/geocoding/${scope}/api?q=` +
      (numero == null ? '' : numero + ' ') +
      address +
      ' ' +
      commune;
    this.logger.debug('Print getCoord ' + encodeURI(req));

    return await lastValueFrom(this.httpService.get(encodeURI(req)));
  }

  /**
   * Set the structure to be deleted in 5 weeks
   */
  public async setToBeDeleted(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    structure.toBeDeletedAt = DateTime.local().plus({ weeks: 5 }).setZone('Europe/Paris').toJSDate();
    structure.save({ timestamps: !hasAdminRole(user) });

    this.sendToBeDeletedNotification(user, structure);
    return structure;
  }

  /**
   * Send an email to structure owners (except the one who asked for the deletion) to inform the structure is will be deleted in 5 weeks
   */
  public async sendToBeDeletedNotification(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureToBeDeleted.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureToBeDeleted.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      toBeDeletedAt: DateTime.fromISO(structure.toBeDeletedAt.toISOString())
        .plus({ days: 1 })
        .toLocaleString(DateTime.DATE_SHORT),
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  /**
   * Cancel the structure deletion when one of the structure owners had asked the structure for deletion
   */
  public async cancelDelete(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    if (!structure.deletedAt) {
      structure.toBeDeletedAt = null;
      structure.save({ timestamps: !hasAdminRole(user) });

      this.sendCancelDeleteNotification(user, structure);
    }
    return structure;
  }

  /**
   * Send an email to other structure owners to inform an owner cancelled the structure deletion
   */
  public async sendCancelDeleteNotification(user: IUser, structure: StructureDocument): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureCancelDelete.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureCancelDelete.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
    });
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    });
  }

  public async deleteOne(user: IUser, structure: StructureDocument): Promise<Structure> {
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    this.structuresSearchService.deleteIndexStructure(structure);
    if (structure.toBeDeletedAt) structure.toBeDeletedAt = null;
    structure.deletedAt = DateTime.local().setZone('Europe/Paris').toJSDate();
    structure.save({ timestamps: !hasAdminRole(user) });
    this.logger.debug(`delete structure : ${structure.structureName} (${structure._id})`);

    // Remove structure from owners (and check if there is a newly unattached user)
    const owners = await this.getOwners(structure._id);
    owners.forEach((owner) => {
      this.userService.removeFromStructureLinked(owner, structure._id);
    });
    this.sendAdminStructureNotification(
      structure,
      this.mailerService.config.templates.structureDeletionNotification.ejs,
      this.mailerService.config.templates.structureDeletionNotification.json
    );
    return structure;
  }

  public async sendAdminNotificationAfterImportingStructure(
    templateLocation: string,
    jsonConfigLocation: string,
    user = null
  ) {
    const uniqueAdminEmails = [...new Set((await this.userService.getAdmins()).map((admin) => admin.email))].map(
      (item) => {
        return { email: item };
      }
    );

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);
    const html = await ejs.renderFile(ejsPath, {
      config,
      user: user,
    });
    this.mailerService.send(uniqueAdminEmails, jsonConfig.subject, html);
  }

  public async sendAdminStructureNotification(
    structure: StructureDocument,
    templateLocation: string,
    jsonConfigLocation: string,
    user = null
  ) {
    const uniqueAdminEmails = [...new Set((await this.userService.getAdmins()).map((admin) => admin.email))].map(
      (item) => {
        return { email: item };
      }
    );

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);
    const html = await ejs.renderFile(ejsPath, {
      config,
      permalink: structure ? structure.permalink : '',
      structureName: structure ? structure.structureName : '',
      structureAddress: structure
        ? `${structure.address.numero || ''} ${structure.address.street} ${structure.address.commune}`
        : '',
      structureDescription: structure ? structure.otherDescription : '',
      user: user,
    });
    this.mailerService.send(uniqueAdminEmails, jsonConfig.subject, html);
  }

  @Cron(CronExpression.EVERY_DAY_AT_4AM)
  public async structuresTasksProcess(): Promise<void> {
    this.logger.log('structuresTasksProcess');

    await this.processToBeDeletedStructures().catch((error) => {
      this.logger.error(error);
    });

    await this.processOutdatedStructures().catch((error) => {
      this.logger.error(error);
    });
  }

  /**
   * Send an email to structure owners and admin in order to accept or decline a join request
   */
  public async sendStructureJoinRequest(
    user: IUser,
    structure: StructureDocument,
    validationToken: string
  ): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureJoinRequest.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureJoinRequest.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      structureName: structure.structureName,
      name: user.name,
      surname: user.surname,
      id: structure._id,
      userId: user._id,
      validationToken: validationToken,
    });
    const owners = await this.getOwners(structure._id);
    for (const owner of owners) {
      if (!owner._id.equals(user._id)) {
        this.mailerService.send(owner.email, jsonConfig.subject, html);
      }
    }
    const admins = await this.userService.getAdmins();
    for (const admin of admins) {
      this.mailerService.send(admin.email, jsonConfig.subject, html);
    }
  }

  public async findWithOwners(idStructure: string): Promise<{ structure: Structure; owners: IUser[] }> {
    const structure = await this.findOne(idStructure);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersDetails(structure._id);
    return { structure: structure, owners: owners };
  }

  public async reportStructureError(structureId: string, content: string) {
    const structure = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersDetails(structure._id);
    const admins = await this.userService.getAdmins();
    const emails = owners.map((owner) => owner.email).concat(admins.map((admin) => admin.email));
    const uniqueEmails = [...new Set(emails)];

    const emailsObject: { email: string }[] = uniqueEmails.map((item) => {
      return { email: item };
    });

    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureErrorReport.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureErrorReport.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      content: content,
      permalink: structure.permalink,
      structureName: structure.structureName,
    });
    this.mailerService.send(emailsObject, jsonConfig.subject, html);
  }

  public async getAllUserCompletedStructures(users: IUser[]) {
    return Promise.all(
      users.map(async (user) => {
        return {
          id: user._id,
          createdAt: user.createdAt ? user.createdAt.toLocaleDateString('fr-FR') : '',
          surname: user.surname,
          name: user.name,
          email: user.email,
          phone: user.phone,
          job: user.job,
          employer: user.employer,
          structures: await Promise.all(
            user.structuresLink.map(async (id) => {
              return this.findOne(id.toHexString());
            })
          ),
        };
      })
    );
  }

  /**
   * Retrieve all structure pending for data.grandlyon consent sharing.
   * Case 1 : null value stands for 'no'
   * Case 2: a date stands for 'yes'
   * Case 3: no key in document when there is no reply to consent
   */
  public async getAllDataConsentPendingStructures(users: IUser): Promise<Structure[]> {
    const data = await Promise.all(
      users.structuresLink.map(async (structureId: Types.ObjectId) => {
        const structure = await this.findOne(structureId.toHexString());
        if (structure && structure.dataShareConsentDate === undefined) {
          return structure;
        }
      })
    );
    return data.filter((value) => value);
  }

  /**
   * Add the personal offer to the structure
   */
  public async addPersonalOffer(
    structureId: string,
    personalOfferDocument: PersonalOfferDocument
  ): Promise<StructureDocument> {
    const structure: StructureDocument = await this.findOne(structureId);
    if (!structure) {
      throw new HttpException('Structure not found for the personal offer attachment', HttpStatus.BAD_REQUEST);
    }
    structure.personalOffers = structure.personalOffers || [];
    if (structure.personalOffers.includes(personalOfferDocument._id)) {
      throw new HttpException('Personal offer already exists in the structure', HttpStatus.BAD_REQUEST);
    } else {
      structure.personalOffers.push(personalOfferDocument._id);
    }
    await structure.save();
    return structure;
  }
  /**
   * Remove a personal offer from structure offers
   */
  public async removePersonalOffer(
    user: IUser,
    structure: StructureDocument,
    offerId: string
  ): Promise<PersonalOfferDocument[]> {
    if (!structure) {
      throw new HttpException('Invalid structure', HttpStatus.NOT_FOUND);
    }
    structure.personalOffers = structure.personalOffers.filter(
      (personalOffer) => !personalOffer._id.equals(new Types.ObjectId(offerId))
    );
    await structure.save({ timestamps: !hasAdminRole(user) });
    return structure.personalOffers;
  }

  public async findByPersonalOfferId(id: string): Promise<StructureDocument> {
    this.logger.debug('findByPersonalOfferId');
    return this.structureModel
      .findOne({
        personalOffers: { $all: [new Types.ObjectId(id)] },
      })
      .populate({ path: 'personalOffers', model: 'PersonalOffer' })
      .exec();
  }

  /**
   * Set structure CTM Territory with data from Data Grand Lyon
   */
  public async setCTMs(): Promise<void> {
    this.logger.debug('setCTMs');
    const structures: StructureDocument[] = await this.findAll();
    for (const structure of structures) {
      const updatedStructure = await this.setCtmTerritory(structure);
      if (updatedStructure) {
        await this.structureModel
          .findByIdAndUpdate(new Types.ObjectId(structure._id), updatedStructure, { timestamps: false })
          .exec();
      }
    }
    return null;
  }

  public async restoreStructure(user: IUser, structureId: string) {
    this.logger.debug(`restoreDeletedStructures: ${structureId}`);
    this.structureModel
      .findByIdAndUpdate(structureId, { $unset: { deletedAt: 1 } }, { timestamps: !hasAdminRole(user) })
      .exec();
  }

  public async updateAllDenormalizedFields() {
    this.logger.log('Update structures denormalized (DN) fields');
    const structures = await this.structureModel.find({ deletedAt: { $exists: false } }).exec();
    structures.forEach((structure) => {
      this.updateDenormalizedFields(structure._id);
    });
    return structures;
  }

  /**
   * Updates the denormalized (DN) fields of the structure
   */
  public async updateDenormalizedFields(idStructure: Types.ObjectId): Promise<StructureDocument> {
    const structure = await this.findOne(idStructure.toHexString());
    if (!structure) {
      throw new HttpException('Invalid structure id', HttpStatus.BAD_REQUEST);
    }
    const owners = await this.userService.getStructureOwnersDetails(structure._id);
    structure.hasNoUserDN = owners.length === 0;

    structure.hasUserWithAppointmentDN = false;
    let conseillerNumFranceServices = false;

    owners.forEach((owner: User) => {
      if (User.getWithAppointment(owner)) {
        // If user has personalOffer for this structure
        const ownerOffers = owner.personalOffers.map((po) => po.id);
        if (structure.personalOffers.filter((structureOffer) => ownerOffers.includes(structureOffer.id)).length) {
          structure.hasUserWithAppointmentDN = true;
        }
      }
      // If user has job 'Conseiller numérique' or 'Conseillère numérique'
      if (owner.job && this.jobsService.isConseillerNumerique(owner.job._id)) {
        conseillerNumFranceServices = true;
      }
    });

    const indexConseillerNum = structure.categories.labelsQualifications.indexOf('conseillerNumFranceServices');
    if (conseillerNumFranceServices && indexConseillerNum === -1) {
      structure.categories.labelsQualifications.push('conseillerNumFranceServices');
    }
    if (!conseillerNumFranceServices && indexConseillerNum > -1) {
      structure.categories.labelsQualifications.splice(indexConseillerNum, 1);
    }

    // markModified is needed otherwise the nested object 'categories' is not saved
    structure.markModified('categories');

    // Setting `timestamps: false` tells Mongoose to skip updating `updatedAt` on this `save()`
    await structure.save({ timestamps: false });

    return structure;
  }

  public async updateLastUpdateMail(structureId: string): Promise<void> {
    this.logger.debug('called UpdateLastUpdateMail');

    const structure = await this.findOne(structureId);
    if (!structure) {
      throw new Error(`Structure with ID ${structureId} not found`);
    }
    const owner = await this.userService.isStructureClaimed(structure._id.toString());
    if (!owner) {
      throw new Error('Structure is unclaimed');
    }
    if (owner.email === undefined) {
      throw new Error(`Structure owner ${owner.name} has no email set`);
    }

    try {
      await this.sendOutdatedEmailToUser(owner.email, structure.structureName, structure._id);
    } catch (error) {
      throw new Error(`Couldn't send email to ${owner.email}. Error: ${error}`);
    }

    await this.structureModel
      .findByIdAndUpdate(structure._id, { lastUpdateMail: new Date() }, { timestamps: false })
      .exec();
  }

  public async touch(structureId: string): Promise<void> {
    const structure = await this.findOne(structureId);
    if (!structure) {
      throw new Error(`Structure with ID ${structureId} not found`);
    }
    await this.structureModel.findByIdAndUpdate(structure._id, structure).exec();
  }

  /**
   * Send an email to prevent outdated
   * a new account.
   */
  public async sendOutdatedEmailToUser(
    userEmail: string,
    structureName: string,
    id: string
  ): Promise<AxiosResponse<unknown>> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(config.templates.structureOutdatedInfo.ejs);
    const jsonConfig = this.mailerService.loadJsonConfig(config.templates.structureOutdatedInfo.json);

    const html = await ejs.renderFile(ejsPath, {
      config,
      name: structureName,
      id: id,
    });
    return this.mailerService.send(userEmail, jsonConfig.subject, html);
  }

  /**
   * Filter offers using structure offers and structure members personalOffers
   */
  private async filterOnOffers(
    structures: StructureDocument[],
    andFiltersOnOffers: Array<any>,
    onlyOffersWithAppointment: boolean
  ): Promise<StructureDocument[]> {
    // Build filters on offers from andFiltersOnOffers
    const filtersOnOffers = { onlineProcedures: [], baseSkills: [], advancedSkills: [] };
    andFiltersOnOffers.forEach((element) => {
      if ('onlineProcedures' in element) filtersOnOffers.onlineProcedures.push(element.onlineProcedures);
      if ('baseSkills' in element) filtersOnOffers.baseSkills.push(element.baseSkills);
      if ('advancedSkills' in element) filtersOnOffers.advancedSkills.push(element.advancedSkills);
    });

    // Set structure offers and social workers personalOffers in structure.categoriesWithPersonalOffers
    const setCategoriesPromises = structures.map((structure) =>
      this.setCategoriesWithPersonalOffers(structure, onlyOffersWithAppointment)
    );
    await Promise.all(setCategoriesPromises);

    // Filter structures on offers checking structure offers and structure social workers offers
    return structures.filter(
      (structure) =>
        filtersOnOffers.onlineProcedures.every((offer) =>
          structure.categoriesWithPersonalOffers.onlineProcedures.includes(offer)
        ) &&
        filtersOnOffers.baseSkills.every((offer) =>
          structure.categoriesWithPersonalOffers.baseSkills.includes(offer)
        ) &&
        filtersOnOffers.advancedSkills.every((offer) =>
          structure.categoriesWithPersonalOffers.advancedSkills.includes(offer)
        )
    );
  }

  /**
   * Parse filter value from string to boolean and map filter category to categories key
   */
  private parseFilter(filters: Array<{ [key: string]: unknown }>): Array<{ [key: string]: unknown }> {
    return filters.map((filter) => {
      const [key, value] = Object.entries(filter)[0];

      if (typeof value === 'string' && value.toLowerCase() === 'true') {
        return { [key]: true };
      } else {
        return { [`categories.${key}`]: value };
      }
    });
  }

  private async findQueryExec(query): Promise<StructureDocument> {
    return query
      .populate({
        path: 'personalOffers',
        model: 'PersonalOffer',
      })
      .populate('structureType')
      .exec();
  }

  /**
   * Set structure position
   */
  private async setStructurePosition(structure: Structure): Promise<Structure> {
    const address = await this.getGeocodingAddress(structure);

    structure.coord = [];
    structure.address.postcode = undefined;
    structure.address.inseeCode = undefined;

    if (address) {
      structure.address.postcode = address.properties.postcode;
      if (address.geometry) {
        structure.coord = address.geometry.coordinates;
      }
    }

    // set territory for CTM filter
    await this.setCtmTerritory(structure);

    return structure;
  }

  /**
   * Get structure photon address
   * (Don't set structure position because it must not be changed in case coord have been set manually in database)
   */
  private async getGeocodingAddress(structure: Structure): Promise<PhotonAddress> {
    let address: PhotonAddress;

    try {
      // Check if address found in photon-bal for this city (photon-bal will found address in another city if city is not in metropole)
      const photonBalCoord = await this.getCoord(
        structure.address.numero || '',
        structure.address.street,
        structure.address.commune,
        'photon-bal'
      );
      address = photonBalCoord.data.features.find(
        (feature) => feature.properties.city?.toLocaleLowerCase() === structure.address.commune.toLocaleLowerCase()
      );

      // If city not found, check Lyon case
      if (!address && structure.address.commune.toLocaleLowerCase().startsWith('lyon')) {
        let cityLyonToSearch = 'lyon';
        // If arrondissement in city, search with arrondissement
        if (structure.address.commune.length > 6) {
          cityLyonToSearch = structure.address.commune.toLocaleLowerCase().substring(0, 6);
        }
        address = photonBalCoord.data.features.find((feature) =>
          feature.properties.city?.toLocaleLowerCase().startsWith(cityLyonToSearch)
        );
      }

      if (!address) {
        // else pick structure from photonban if it exists or log error
        const photonCoord = await this.getCoord(
          structure.address.numero || '',
          structure.address.street,
          structure.address.commune,
          'photon'
        );
        address = photonCoord.data.features.find((feature) => feature.properties.hasOwnProperty('housenumber'));
        // If not found and no number ("Place Pierre Marie Durand L'Arbresle" for example), get the first found
        if (!address && !structure.address.numero && photonCoord.data.features.length) {
          address = photonCoord.data.features[0];
        }
        // If found in another city ("40 Place Laurent Bonnevay Villefranche-sur-Saône" for example), find first item with the right city (but without housenumber)
        if (address?.properties.city?.toLocaleLowerCase() !== structure.address.commune.toLocaleLowerCase()) {
          // curl https://download.data.grandlyon.com/geocoding/photon/api?q=230%20Route%20du%20Dauphin%C3%A9%20Colombier-Saugnieu returns "city":"Colombier"...
          if (
            !(
              address.properties.city?.toLocaleLowerCase() === 'colombier' &&
              structure.address.commune.toLocaleLowerCase() === 'colombier-saugnieu'
            )
          ) {
            address = photonCoord.data.features.find(
              (feature) =>
                feature.properties.city?.toLocaleLowerCase() === structure.address.commune.toLocaleLowerCase()
            );
          }
        }
      }

      if (!address) {
        this.logger.error(
          `No photon address found for: "${structure.address.numero || ''} ${structure.address.street} ${
            structure.address.commune
          }"`
        );
      } else {
        if (!address.geometry) {
          this.logger.error(
            `No coord found for: "${structure.address.numero || ''} ${structure.address.street} ${
              structure.address.commune
            }"`
          );
        }
        if (!address.properties.postcode) {
          this.logger.error(
            `No postcode found for structure "${structure.structureName}" (${JSON.stringify(
              structure.address
            )} and ${JSON.stringify(address)})`
          );
        }
      }
    } catch (err) {
      this.logger.error(err);
    }

    return address;
  }

  /**
   * Get structure territory
   */
  private async setCtmTerritory(structure: Structure): Promise<Structure> {
    structure.categories['ctm'] = [];

    if (!structure.address) {
      this.logger.error(`No address for structure "${structure.structureName}"`);
      return structure;
    }

    if (!structure.address.postcode) {
      const address = await this.getGeocodingAddress(structure);
      if (address) {
        structure.address.postcode = address.properties.postcode;
      } else {
        this.logger.error(`No photon address found for structure "${structure.structureName}"`);
        return structure;
      }
    }

    if (structure.address.postcode) {
      const inseeCode = await this.getInseeCode(structure.address.postcode, structure.address.commune);
      if (inseeCode) {
        structure.address.inseeCode = inseeCode;
        const ctm: Module = await this.categoriesService.findCtmByInseeCode(inseeCode);
        if (ctm) {
          this.logger.log(`${structure.structureName} : CTM ${ctm.id}`);
          structure.categories['ctm'] = [ctm.id];
        }
      }
    }
    return structure;
  }

  private async processToBeDeletedStructures(): Promise<void> {
    this.logger.log('processToBeDeletedStructures');

    const structures = await this.structureModel
      .find()
      .where('toBeDeletedAt')
      .lte(DateTime.local().toMillis())
      .where('deletedAt')
      .exists(false)
      .exec();

    const pendingStructures = (await this.userService.getPendingStructures()) as PendingStructureDto[];
    const structureIds = pendingStructures.map((pending) => pending.structureId);
    structures.forEach((structure) => {
      if (structureIds.includes(structure.id)) {
        this.logger.debug(`cancel structure soft-delete : ${structure.structureName} (${structure._id})`);
        this.structureModel
          .findByIdAndUpdate(
            new Types.ObjectId(structure.id),
            {
              toBeDeletedAt: null,
            },
            {
              timestamps: false,
            }
          )
          .exec();
      } else {
        this.logger.log(`delete structure : ${structure.structureName} (${structure._id})`);
        this.deleteOne(null, structure);
      }
    });
  }

  private async processOutdatedStructures(): Promise<void> {
    this.logger.debug('processOutdatedStructures');
    const OUTDATED_MONTH_TO_CHECK = 6;
    const structureList = await this.findAll();
    // Get outdated structures
    const filteredList = structureList.filter((structure) => {
      const updatedDate = DateTime.fromISO(structure.updatedAt.toISOString());
      const diff = updatedDate.diffNow('month').months;
      if (Math.abs(diff) > OUTDATED_MONTH_TO_CHECK) {
        this.logger.debug(`found outdated structure : ${structure.structureName}`);
        return true;
      }
    });
    // Get owners of outdated structures
    const ownerList = await Promise.all(
      filteredList.map(async (structure) => {
        const owner = await this.userService.isStructureClaimed(structure._id.toString());
        if (owner) {
          return { structure: structure, owner: owner };
        }
      })
    );
    // Send email if possible and update user
    ownerList
      .filter((x) => x != undefined)
      .forEach(async (data) => {
        const user = await this.userService.findOne(data.owner.email);
        // If mail is already sent, do not resend
        if (user.structureOutdatedMailSent.includes(data.structure._id)) {
          return;
        } else {
          try {
            await this.sendOutdatedEmailToUser(data.owner.email, data.structure.structureName, data.structure._id);
            user.structureOutdatedMailSent.push(data.structure._id);
            user.save();
          } catch (error) {
            this.logger.error(`Couldn't send email to ${data.owner.email}. Error: ${error}`);
          }
        }
      });
  }

  private async getOwners(structureId: string): Promise<IUser[]> {
    // Get owners of outdated structures
    return this.userService.getStructureOwners(structureId);
  }
}
