import { HttpService } from '@nestjs/axios';
import {
  Body,
  Controller,
  Delete,
  Get,
  HttpException,
  HttpStatus,
  Logger,
  Param,
  Patch,
  Post,
  Query,
  Request,
  UseGuards,
} from '@nestjs/common';
import { AuthGuard } from '@nestjs/passport';
import { ApiOperation, ApiParam, ApiResponse, ApiTags } from '@nestjs/swagger';
import * as _ from 'lodash';
import { Types } from 'mongoose';
import { JwtAuthGuard } from '../auth/guards/jwt-auth.guard';
import { CategoriesService } from '../categories/services/categories.service';
import { hasAdminRole } from '../shared/utils';
import { CreateTempUserDto } from '../temp-user/dto/create-temp-user.dto';
import { ITempUser } from '../temp-user/temp-user.interface';
import { TempUser } from '../temp-user/temp-user.schema';
import { TempUserService } from '../temp-user/temp-user.service';
import { Roles } from '../users/decorators/roles.decorator';
import { IsStructureOwnerGuard } from '../users/guards/isStructureOwner.guard';
import { RolesGuard } from '../users/guards/roles.guard';
import { pendingStructuresLink } from '../users/interfaces/pendingStructure';
import { IUser } from '../users/interfaces/user.interface';
import { User } from '../users/schemas/user.schema';
import { UsersService } from '../users/services/users.service';
import { PersonalOffersService } from './../personal-offers/personal-offers.service';
import { depRegex } from './common/regex';
import { CreateStructureDto } from './dto/create-structure.dto';
import { QueryStructure } from './dto/query-structure.dto';
import { UpdateStructureDto } from './dto/update-structure.dto';
import { PhotonPoints } from './interfaces/photon-response.interface';
import { StructureFormatted } from './interfaces/structure-formatted.interface';
import { Structure, StructureDocument } from './schemas/structure.schema';
import { StructuresExportService } from './services/structures-export.service';
import { StructuresService } from './services/structures.service';
import { lastValueFrom } from 'rxjs';

@ApiTags('structures')
@Controller('structures')
export class StructuresController {
  private readonly logger = new Logger(StructuresController.name);
  constructor(
    private readonly httpService: HttpService,
    private readonly structureService: StructuresService,
    private readonly structuresExportService: StructuresExportService,
    private readonly userService: UsersService,
    private readonly tempUserService: TempUserService,
    private readonly categoriesService: CategoriesService,
    private readonly personalOffersService: PersonalOffersService
  ) {}

  /**
   * Return points of given town exist.
   * @returns Array of points
   */
  @Get('coordinates/:zipcode')
  @ApiParam({ name: 'zipcode', type: String, required: true })
  public async getCoordinates(@Param('zipcode') city: string): Promise<number[][]> {
    const extent = await lastValueFrom(
      this.httpService.get(encodeURI(`https://download.data.grandlyon.com/geocoding/photon/api`), {
        params: { q: `${city}`, layer: 'city' },
      })
    )
      .then(async (res) => res.data.features)
      .then((data) =>
        data.filter(
          (cityPoint: PhotonPoints) =>
            /* check if query string = postcode 
                     OR query string = city (without hyphens and diacritics)
                     OR query string = name (in 2 cases, Charly et Rochetaillée-sur-saône, the city name is in the name field and the city field is missing) */
            (cityPoint.properties.postcode == city ||
              _.deburr(cityPoint.properties.city?.toLowerCase().replace(/-/g, ' ')).includes(
                _.deburr(city).toLowerCase().replace(/-/g, ' ')
              ) ||
              _.deburr(cityPoint.properties.name?.toLowerCase().replace(/-/g, ' ')) ===
                _.deburr(city).toLowerCase().replace(/-/g, ' ')) &&
            // check result has postcode 69 (except for Lyon which doesn't have postcode field)
            (cityPoint.properties.postcode?.match(depRegex) || cityPoint.properties.name == 'Lyon')
        )
      )
      .then((data) => data.map((filteredCityPoint) => filteredCityPoint.properties.extent));

    // If the city extent is found, convert it into 2 map bounds points
    if (extent.length) {
      return [
        [extent[0][0], extent[0][1]],
        [extent[0][2], extent[0][3]],
      ];
    } else {
      return [];
    }
  }

  @Post()
  public async create(@Body() createStructureDto: CreateStructureDto): Promise<Structure> {
    Logger.debug(`create | ${createStructureDto.structure.structureName}`, StructuresController.name);
    return this.structureService.create(createStructureDto.userEmail, createStructureDto.structure);
  }

  @Post('search')
  public async search(@Query() query: QueryStructure, @Body() body): Promise<Structure[]> {
    this.logger.debug(`search`);
    return this.structureService.searchForStructures(
      query.query,
      body ? body.filters : null,
      null,
      body?.onlyOffersWithAppointment || false,
      body?.limit || null,
      body?.inseeCodes || null
    );
  }

  @Post('searchByName')
  public async searchByName(@Query() query: QueryStructure, @Body() body): Promise<Structure[]> {
    this.logger.debug('searchByName');
    return this.structureService.searchForStructures(query.query, body ? body.filters : null, ['structureName']);
  }

  @Post('resetSearchIndex')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @Roles('admin')
  public async resetES(): Promise<StructureDocument[]> {
    // Update denormalized fields (which are used in search)
    await this.structureService.updateAllDenormalizedFields();
    return await this.structureService.initiateStructureIndex();
  }

  @Post('updateDN')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @Roles('admin')
  public async updateDN(): Promise<StructureDocument[]> {
    // Update denormalized fields
    return await this.structureService.updateAllDenormalizedFields();
  }

  @ApiOperation({ summary: 'Patch one value or more of a structure' })
  @Patch(':id')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @Roles('admin')
  public async update(@Param('id') id: string, @Body() body: UpdateStructureDto, @Request() req): Promise<Structure> {
    return this.structureService.update(id, body, req.user);
  }

  @Get()
  public async findAll(): Promise<Structure[]> {
    this.logger.debug('findAll');
    return this.structureService.findAll();
  }

  @Get('formated')
  public async findAllFormatted(): Promise<StructureFormatted[]> {
    this.logger.debug('findAllFormatted');
    return this.structuresExportService.exportFormatted();
  }

  @Post(':id/claim')
  public async claim(@Param('id') idStructure: string, @Body() user: User): Promise<pendingStructuresLink[]> {
    this.logger.debug(`claim structure ${idStructure} from ${user.email}`);
    const structure = await this.structureService.findOne(idStructure);
    return this.userService.updateStructureLinkedClaim(user.email, idStructure, structure);
  }

  @Post('address')
  public async searchAddress(@Body() data: { searchQuery: string }) {
    return this.structureService.searchAddress(data);
  }

  @Get(':id')
  @UseGuards(AuthGuard(['jwt', 'anonymous'])) // first success wins (allow anonymous call, req.user is an empty object if user is not authenticated)
  public async find(@Request() req, @Param('id') id: string) {
    this.logger.debug(`find with ${id}`);
    const structure = await this.structureService.findOne(id);
    if (!structure || (structure.deletedAt && !hasAdminRole(req.user))) {
      this.logger.log(`structure with ${id} does not exist`);
      throw new HttpException('Structure does not exist', HttpStatus.NOT_FOUND);
    }
    return structure;
  }

  @Get('permalink/:permalink')
  @UseGuards(AuthGuard(['jwt', 'anonymous'])) // first success wins (allow anonymous call, req.user is an empty object if user is not authenticated)
  public async findByPermalink(@Request() req, @Param('permalink') permalink: string) {
    this.logger.debug(`findByPermalink with ${permalink}`);
    const structure = await this.structureService.findByPermalink(permalink);
    if (!structure || (structure.deletedAt && !hasAdminRole(req.user))) {
      this.logger.log(`structure with permalink ${permalink} does not exist`);
      throw new HttpException('Structure does not exist', HttpStatus.NOT_FOUND);
    }
    return structure;
  }

  @Get(':id/withOwners')
  public async findWithOwners(@Param('id') id: string) {
    return this.structureService.findWithOwners(id);
  }

  @Get(':id/tempUsers')
  public async getTempUsers(@Param('id') id: string) {
    return this.tempUserService.getStructureTempUsers(id);
  }

  @ApiOperation({
    description: `Mettre à jour les territoires CTM des structures à partir de Data Grand Lyon`,
  })
  @ApiResponse({
    status: 201,
    description: 'The CTM territories have been updated successfully.',
  })
  @Post('/ctm/update')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @Roles('admin')
  public async updateCTM(): Promise<void> {
    this.logger.log('updateCTM');
    await this.categoriesService.updateCTM();
    await this.structureService.setCTMs();
    this.logger.log('updateCTM done');
    return;
  }

  @Delete(':id')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  public async delete(@Request() req, @Param('id') id: string) {
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    const otherOwners: IUser[] = (await this.userService.getStructureOwners(id)).filter((owner) => {
      return !owner._id.equals(req.user._id);
    });
    if (otherOwners.length === 0 || hasAdminRole(req.user)) {
      return this.structureService.deleteOne(req.user, structure);
    } else {
      return this.structureService.setToBeDeleted(req.user, structure);
    }
  }

  @Post(':id/cancelDelete')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  public async cancelDelete(@Request() req, @Param('id') id: string): Promise<Structure> {
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    return this.structureService.cancelDelete(req.user, structure);
  }

  @Post(':id/addOwner')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @Roles('admin')
  @ApiParam({ name: 'id', type: String, required: true })
  public async addOwner(
    @Param('id') id: string,
    @Body() user: CreateTempUserDto
  ): Promise<Types.ObjectId[] | TempUser> {
    this.logger.debug(`addOwner`);
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    user.structuresLink = [new Types.ObjectId(id)];

    // If user already exist, use created account
    if (await this.userService.verifyUserExist(user.email)) {
      this.tempUserService.sendUserMail(user as ITempUser, structure.structureName, true);
      return this.userService.updateStructureLinked(user.email, id);
    }
    // If temp user exists, update it
    const existingTempUser = await this.tempUserService.findOne(user.email);
    if (existingTempUser) {
      user['_id'] = existingTempUser._id;
      this.tempUserService.sendUserMail(user as ITempUser, structure.structureName);
      return this.tempUserService.updateStructureLinked(user);
    }
    // If not, create
    return this.tempUserService.create(user, structure.structureName);
  }

  @Delete(':id/owner/:userId')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  @ApiParam({ name: 'userId', type: String, required: true })
  public async removeOwner(@Request() req, @Param('id') id: string, @Param('userId') userId: string): Promise<void> {
    this.logger.debug(`removeOwner`);
    // Get structure
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get user
    const userFromDb = await this.userService.findById(userId);
    if (!userFromDb || !userFromDb.structuresLink.some((structure) => structure.equals(id))) {
      throw new HttpException('Invalid User', HttpStatus.NOT_FOUND);
    }
    await this.userService.removeFromStructureLinked(userFromDb, id);

    // If user has personal offer, delete it and remove it from both user and structure offers
    const personalOffer = this.userService.getPersonalOfferInStructure(userFromDb, structure);
    if (personalOffer) {
      this.personalOffersService.delete(req.user, personalOffer._id);
    }
  }

  @Delete(':id/tempUser/:userId')
  @UseGuards(JwtAuthGuard, IsStructureOwnerGuard)
  @ApiParam({ name: 'id', type: String, required: true })
  @ApiParam({ name: 'userId', type: String, required: true })
  public async removeTempUser(@Param('id') id: string, @Param('userId') userId: string): Promise<void> {
    // Get structure
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get temp user
    const userFromDb = await this.tempUserService.findById(userId);
    if (!userFromDb || !userFromDb.structuresLink.includes(new Types.ObjectId(id))) {
      throw new HttpException('Invalid temp user', HttpStatus.NOT_FOUND);
    }
    this.tempUserService.removeFromStructureLinked(userFromDb.email, id);
  }

  @Post('reportStructureError')
  public async reportStructureError(@Body() data: { structureId: string; content: string }): Promise<void> {
    return this.structureService.reportStructureError(data.structureId, data.content);
  }
}
