import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test } from '@nestjs/testing';
import * as ejs from 'ejs';
import { MailerModule } from '../mailer/mailer.module';
import { MailerService } from '../mailer/mailer.service';
import { TempUserService } from './temp-user.service';
import { Types } from 'mongoose';
import { CreateTempUserDto } from './dto/create-temp-user.dto';

const ejsSpy = jest.spyOn(ejs, 'renderFile');

describe('TempUserService', () => {
  let tempUserService: TempUserService;

  const tempUserModelMock = {
    create: jest.fn(),
    findOne: jest.fn().mockReturnThis(),
    findById: jest.fn(),
    findByIdAndUpdate: jest.fn(),
    deleteOne: jest.fn().mockReturnThis(),
    find: jest.fn().mockReturnThis(),
    exec: jest.fn(),
    updateOne: jest.fn(),
    populate: jest.fn().mockReturnThis(),
    select: jest.fn(),
  };

  const mockMailService = {
    send: jest.fn(),
    getTemplateLocation: jest.fn(),
    loadJsonConfig: jest.fn().mockReturnValue({ subject: 'subject' }),
    config: {
      templates: {
        userAddedToStructure: {
          ejs: 'userAddedToStructure.ejs',
          json: 'userAddedToStructure.json',
        },
        tempUserRegistration: {
          ejs: 'tempUserRegistration.ejs',
          json: 'tempUserRegistration.json',
        },
      },
    },
  };

  beforeEach(async () => {
    const module = await Test.createTestingModule({
      imports: [HttpModule, MailerModule],
      providers: [
        TempUserService,
        {
          provide: getModelToken('TempUser'),
          useValue: tempUserModelMock,
        },
        { provide: MailerService, useValue: mockMailService },
      ],
    }).compile();

    tempUserService = module.get<TempUserService>(TempUserService);
  });

  it('should be defined', () => {
    expect(tempUserService).toBeDefined();
  });

  describe('findAll', () => {
    it('should return a list of temp users', async () => {
      const tempUsers = [
        { email: 'user1@test.com', structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')] },
        { email: 'user2@test.com', structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')] },
      ];
      tempUserModelMock.exec.mockResolvedValueOnce(tempUsers);
      expect(await tempUserService.findAll()).toEqual(tempUsers);
    });
    it('should throw and error', async () => {
      tempUserModelMock.exec.mockRejectedValueOnce(new Error('test'));
      try {
        await tempUserService.findAll();
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('An error occurred while fetching temp users.');
        expect(e.status).toEqual(HttpStatus.INTERNAL_SERVER_ERROR);
      }
    });
  });

  describe('create', () => {
    const tmpUser: CreateTempUserDto = {
      email: 'test@test.com',
      structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
    };

    it('should not create temporary user if it already exists', async () => {
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      try {
        await tempUserService.create(tmpUser, 'PIMMS Framboise');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('User already exists');
        expect(e.status).toEqual(HttpStatus.BAD_REQUEST);
      }
    });

    it('should create a temporary user', async () => {
      tempUserModelMock.exec.mockResolvedValueOnce(null);
      tempUserModelMock.create.mockResolvedValueOnce(tmpUser);
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      ejsSpy.mockResolvedValue('mock');
      expect(await tempUserService.create(tmpUser, 'PIMMS Framboise')).toEqual(tmpUser);
      expect(mockMailService.send).toHaveBeenCalled();
    });
  });

  it('should find one', async () => {
    const tmpUser = {
      email: 'test2@test.com',
      structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
    };
    tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
    expect(await tempUserService.findOne('test2@test.com')).toEqual(tmpUser);
  });

  it('should find one by id', async () => {
    const tmpUser = {
      email: 'test2@test.com',
      structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
    };
    tempUserModelMock.findById.mockReturnThis();
    tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
    expect(await tempUserService.findById('5fbb92e480a5c257dc0161f0')).toEqual(tmpUser);
  });

  describe('delete', () => {
    it('should delete a temporary user', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
      };
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      tempUserModelMock.exec.mockImplementationOnce(() => ({}));
      expect(await tempUserService.delete('toto@test.com')).toEqual(tmpUser);
    });

    it('should return an error if the user does not exist', async () => {
      tempUserModelMock.exec.mockResolvedValueOnce(null);
      try {
        await tempUserService.delete('toto@test.com');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('User does not exist');
        expect(e.status).toEqual(HttpStatus.BAD_REQUEST);
      }
    });
  });

  describe('updateStructureLinked', () => {
    it('should not update the linked structure if the user is not found', async () => {
      const createTempUserDto = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
      };
      tempUserModelMock.exec.mockResolvedValueOnce(null);

      try {
        await tempUserService.updateStructureLinked(createTempUserDto);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('User not found');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });

    it('should not update the linked structure if the user is already linked', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: ['653015f722f6c63019ab038d'],
      } as any;
      const existingUserInDb = {
        structuresLink: ['653015f722f6c63019ab038d', '653015f722f6c63019ab038e'],
        save: jest.fn().mockResolvedValueOnce(tmpUser),
      };

      tempUserModelMock.exec.mockResolvedValueOnce(existingUserInDb);

      try {
        await tempUserService.updateStructureLinked(tmpUser);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('User already linked');
        expect(e.status).toEqual(HttpStatus.UNPROCESSABLE_ENTITY);
      }
    });
    it('should update the linked structure', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('653015f722f6c63019ab038d')],
      };
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);

      const userInDb = {
        structuresLink: [],
        save: jest.fn().mockResolvedValueOnce(tmpUser),
      };

      jest.spyOn(tempUserModelMock, 'findOne').mockReturnValue({
        exec: jest.fn().mockResolvedValueOnce(userInDb),
      });

      tempUserModelMock.findByIdAndUpdate.mockReturnThis();
      expect(await tempUserService.updateStructureLinked(tmpUser)).toEqual(tmpUser);
    });
  });

  describe('getStructureTempUsers', () => {
    it('should return a list of temp users', async () => {
      tempUserModelMock.select.mockReturnThis();
      await tempUserService.getStructureTempUsers('653015f722f6c63019ab038d');
      expect(tempUserModelMock.find).toHaveBeenCalled();
    });
  });

  describe('removeFromStructureLinked', () => {
    it('should not find the user', async () => {
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(null);
      try {
        await tempUserService.removeFromStructureLinked('jp@test.com', '653015f722f6c63019ab038d');
      } catch (e) {
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
        expect(e.message).toEqual('Invalid temp user');
      }
    });
    it('should fail because temp user is not linked to the structure', async () => {
      const tmpUser = {
        email: 'test2@test.com',
        structuresLink: [new Types.ObjectId('aaa015f722f6c63019ab038d')],
      };
      tempUserModelMock.findOne.mockReturnThis();
      tempUserModelMock.exec.mockResolvedValueOnce(tmpUser);
      try {
        await tempUserService.removeFromStructureLinked('jp@test.com', '653015f722f6c63019ab038d');
      } catch (e) {
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
        expect(e.message).toEqual("Temp user doesn't belong to this structure");
      }
    });
  });
});
