import {
  Body,
  Controller,
  Delete,
  Get,
  HttpException,
  HttpStatus,
  Logger,
  Param,
  Post,
  Put,
  Query,
  Request,
  UseGuards,
} from '@nestjs/common';
import { ApiBearerAuth, ApiOperation, ApiParam, ApiTags } from '@nestjs/swagger';
import { MergeJobDto } from '../../admin/dto/merge-job.dto';
import { JwtAuthGuard } from '../../auth/guards/jwt-auth.guard';
import { hasAdminRole } from '../../shared/utils';
import { Roles } from '../decorators/roles.decorator';
import { CreateJobDto } from '../dto/create-job.dto';
import { RolesGuard } from '../guards/roles.guard';
import { Job } from '../schemas/job.schema';
import { JobsService } from '../services/jobs.service';
import { UsersService } from '../services/users.service';
import { UpdateJobDto } from '../dto/update-job.dto';

@ApiTags('jobs')
@Controller('jobs')
export class JobsController {
  private readonly logger = new Logger(JobsController.name);

  constructor(
    private jobsService: JobsService,
    private usersService: UsersService
  ) {}

  /**
   * Return every jobs
   */
  @Get()
  public async findAll(@Query() query?: { search: string }) {
    this.logger.debug('findAll ' + (query?.search ? `with query ${query.search}` : 'without query'));
    return this.jobsService.findAll(query?.search);
  }

  /**
   * Create a new job
   */
  @Post()
  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @ApiParam({ name: 'job', type: CreateJobDto, required: true })
  public async createJob(@Request() req, @Body() job: CreateJobDto): Promise<Job> {
    this.logger.debug(`createJob: ${job.name}`);
    const existingJob = await this.jobsService.findByName(job.name);
    if (existingJob) {
      return Promise.resolve(existingJob);
    }
    // if user is admin, do not send notification
    if (hasAdminRole(req.user)) {
      return this.jobsService.create(job, true, job.hasPersonalOffer, false);
    }
    return this.jobsService.create(job);
  }

  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @Get('validated')
  @ApiOperation({ description: 'Get validated jobs populated with users attached to it' })
  public async findValidatedJobs() {
    const jobs = await this.jobsService.findAll();
    return this.usersService.populateJobsWithUsers(jobs);
  }

  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @Get('unvalidated')
  @ApiOperation({ description: 'Get unvalidated jobs populated with users attached to it' })
  public async findUnvalidatedJobs() {
    const jobs = await this.jobsService.findAllUnvalidated();
    return this.usersService.populateJobsWithUsers(jobs);
  }

  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @Post('validate/:id')
  @ApiParam({ name: 'id', type: String, required: true })
  @ApiOperation({ description: 'Validate job' })
  public async validateJob(@Param() params) {
    return this.jobsService.validate(params.id);
  }

  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @Put('merge')
  public async mergeJob(@Body() mergeJob: MergeJobDto) {
    this.logger.debug('mergeJob');
    return this.jobsService.mergeJob(mergeJob);
  }

  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @Put(':id')
  public async updateJob(@Param('id') id: string, @Body() body: UpdateJobDto) {
    this.logger.debug('updateJob');
    return this.jobsService.update(id, body);
  }

  @UseGuards(JwtAuthGuard, RolesGuard)
  @Roles('admin')
  @Delete(':id')
  @ApiBearerAuth('JWT')
  @ApiParam({ name: 'id', type: String, required: true })
  public async deleteJob(@Param() params) {
    const researchedJob = await this.jobsService.findOne(params.id);
    // look for any relations within user collection, reject action if found
    if (researchedJob === null) {
      throw new HttpException('Job does not exists', HttpStatus.NOT_FOUND);
    }
    const isJobLinked = await this.usersService.isJobLinkedToUser(researchedJob._id);
    if (isJobLinked) {
      throw new HttpException('Cannot delete job. It has user(s) attached to it.', HttpStatus.FORBIDDEN);
    }
    return this.jobsService.deleteOneId(params.id);
  }
}
