import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test } from '@nestjs/testing';
import { Types } from 'mongoose';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { CreateJobsGroupsDto } from '../dto/create-jobsGroups.dto';
import { JobsGroups } from '../schemas/jobsGroups.schema';
import { JobsService } from '../services/jobs.service';
import { JobsGroupsService } from '../services/jobsGroups.service';
import { JobsGroupsController } from './jobsGroups.controller';

describe('JobsGroupsController', () => {
  let jobsGroupsController: JobsGroupsController;

  const jobsGroupsServiceMock = {
    findAll: jest.fn(),
    findByName: jest.fn(),
    create: jest.fn(),
    findOne: jest.fn(),
    deleteOne: jest.fn(),
    update: jest.fn(),
    isUsed: jest.fn(),
  };

  const jobServiceMock = {
    getJobsGroupsAttachedJobs: jest.fn(),
  };

  beforeEach(async () => {
    const module = await Test.createTestingModule({
      imports: [ConfigurationModule, HttpModule],
      providers: [
        {
          provide: JobsGroupsService,
          useValue: jobsGroupsServiceMock,
        },
        {
          provide: JobsService,
          useValue: jobServiceMock,
        },
        {
          provide: getModelToken('JobsGroups'),
          useValue: JobsGroups,
        },
      ],
      controllers: [JobsGroupsController],
    }).compile();

    jobsGroupsController = module.get<JobsGroupsController>(JobsGroupsController);
  });

  it('should be defined', () => {
    expect(jobsGroupsController).toBeDefined();
  });

  describe('findAll', () => {
    it('should findAll', async () => {
      await jobsGroupsController.findAll();
      expect(jobsGroupsServiceMock.findAll).toHaveBeenCalled();
    });
  });

  describe('findAllWithJobs', () => {
    it('should findAll with jobs', async () => {
      await jobsGroupsController.findAllWithJobs();
      expect(jobsGroupsServiceMock.findAll).toHaveBeenCalledWith(true);
    });
  });

  describe('createJobsGroups', () => {
    it('should throw error on already existing job group `Dev`', async () => {
      const newJobsGroups: CreateJobsGroupsDto = {
        name: 'Dev',
      };
      jobsGroupsServiceMock.findByName.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5a7'),
        name: 'Dev',
      });
      try {
        await jobsGroupsController.createJobsGroups(newJobsGroups);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Job group already exist');
        expect(e.status).toBe(HttpStatus.UNPROCESSABLE_ENTITY);
      }
    });

    it('should create job group `Dev`', async () => {
      const newJobsGroups: CreateJobsGroupsDto = {
        name: 'Dev',
      };
      const newCreatedJobsGroups: JobsGroups = {
        name: 'Dev',
      };
      jobsGroupsServiceMock.findByName.mockResolvedValueOnce(null);
      jobsGroupsServiceMock.create.mockResolvedValueOnce(newCreatedJobsGroups);
      await jobsGroupsController.createJobsGroups(newJobsGroups);
      expect(jobsGroupsServiceMock.create).toHaveBeenCalled();
    });
  });

  describe('Edit Job Group', () => {
    it('should update job group', async () => {
      jobsGroupsServiceMock.update.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'DevMod',
      });
      expect(await jobsGroupsController.updateJobsGroups('6231aefe76598527c8d0b5bc', { name: 'DevMod' })).toBeTruthy();
    });
  });
  describe('Delete Job Group', () => {
    jobsGroupsServiceMock.deleteOne.mockResolvedValueOnce({
      n: 1,
      ok: 1,
      deletedCount: 1,
    });

    it('should not be able to find job group', async () => {
      jobsGroupsServiceMock.findOne.mockResolvedValueOnce(null);
      try {
        await jobsGroupsController.deleteJobsGroups({ id: '6231aefe76598527c8d0b5ba' });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Job group not found');
        expect(e.status).toEqual(HttpStatus.NOT_FOUND);
      }
    });

    it('should not delete job group if a job is linked', async () => {
      jobsGroupsServiceMock.findOne.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Dev',
      });
      jobServiceMock.getJobsGroupsAttachedJobs.mockResolvedValueOnce([{ test: 1 }]);
      try {
        await jobsGroupsController.deleteJobsGroups({ id: '6231aefe76598527c8d0b5bc' });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toEqual('Cannot delete job group. It has job(s) attached to it.');
        expect(e.status).toEqual(HttpStatus.FORBIDDEN);
      }
    });

    it('should delete job group', async () => {
      jobsGroupsServiceMock.findOne.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5ba'),
        name: 'Dev',
      });
      jobServiceMock.getJobsGroupsAttachedJobs.mockResolvedValueOnce([]);
      await jobsGroupsController.deleteJobsGroups({ id: '6231aefe76598527c8d0b5ba' });
      expect(jobsGroupsServiceMock.deleteOne).toHaveBeenCalled();
    });
  });
});
