import {
  Body,
  Controller,
  Delete,
  Get,
  HttpException,
  HttpStatus,
  Logger,
  Param,
  Post,
  Put,
  UseGuards,
} from '@nestjs/common';
import { ApiBearerAuth, ApiParam, ApiTags } from '@nestjs/swagger';
import { JwtAuthGuard } from '../../auth/guards/jwt-auth.guard';
import { Roles } from '../decorators/roles.decorator';
import { CreateJobsGroupsDto } from '../dto/create-jobsGroups.dto';
import { RolesGuard } from '../guards/roles.guard';
import { JobsGroups } from '../schemas/jobsGroups.schema';
import { JobsGroupsService } from '../services/jobsGroups.service';
import { JobsService } from '../services/jobs.service';

@ApiTags('jobs-groups')
@Controller('jobs-groups')
export class JobsGroupsController {
  private readonly logger = new Logger(JobsGroupsController.name);

  constructor(
    private jobsGroupsService: JobsGroupsService,
    private jobsService: JobsService
  ) {}

  /**
   * Return every jobsGroups
   */
  @Get()
  public async findAll() {
    this.logger.debug('findAll JobsGroups');
    return this.jobsGroupsService.findAll(false);
  }

  /**
   * Return every jobsGroups
   */
  @Get('with-jobs')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  public async findAllWithJobs() {
    this.logger.debug('findAll JobsGroups with jobs');
    return this.jobsGroupsService.findAll(true);
  }

  /**
   * Create a new jobsGroups
   */
  @Post()
  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @ApiParam({ name: 'jobsGroups', type: CreateJobsGroupsDto, required: true })
  public async createJobsGroups(@Body() jobsGroups: CreateJobsGroupsDto): Promise<JobsGroups> {
    this.logger.debug(`createJobsGroups: ${jobsGroups.name}`);
    const existingJobsGroups = await this.jobsGroupsService.findByName(jobsGroups.name);
    if (existingJobsGroups) {
      this.logger.warn(`Job group already exist: ${jobsGroups.name}`);
      throw new HttpException('Job group already exist', HttpStatus.UNPROCESSABLE_ENTITY);
    }
    return this.jobsGroupsService.create(jobsGroups);
  }

  @Put(':id')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  public async updateJobsGroups(@Param('id') id: string, @Body() body: CreateJobsGroupsDto): Promise<JobsGroups> {
    this.logger.debug('updateJobsGroups');
    return this.jobsGroupsService.update(id, body);
  }

  @Delete(':id')
  @UseGuards(JwtAuthGuard, RolesGuard)
  @ApiBearerAuth('JWT')
  @Roles('admin')
  @ApiParam({ name: 'id', type: String, required: true })
  public async deleteJobsGroups(@Param() params): Promise<JobsGroups> {
    const jobsGroups = await this.jobsGroupsService.findOne(params.id);
    if (!jobsGroups) {
      throw new HttpException('Job group not found', HttpStatus.NOT_FOUND);
    }
    // look for any relations within job collection, reject action if found
    const jobsLinked = await this.jobsService.getJobsGroupsAttachedJobs(jobsGroups._id);
    if (jobsLinked?.length) {
      throw new HttpException('Cannot delete job group. It has job(s) attached to it.', HttpStatus.FORBIDDEN);
    }
    return this.jobsGroupsService.deleteOne(jobsGroups);
  }
}
