import { getModelToken } from '@nestjs/mongoose';
import { Test } from '@nestjs/testing';
import { multipleUsers } from '../../../test/mock/data/users.mock.data';
import { mockJwtAuthGuard } from '../../../test/mock/guards/jwt-auth.mock.guard';
import { mockRoleGuard } from '../../../test/mock/guards/role.mock.guard';
import { JwtAuthGuard } from '../../auth/guards/jwt-auth.guard';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { RolesGuard } from '../guards/roles.guard';
import { User } from '../schemas/user.schema';
import { UserRegistryService } from '../services/userRegistry.service';
import { UsersRegistryController } from './userRegistry.controller';

describe('UserRegistryController', () => {
  let userRegistryController: UsersRegistryController;

  const userRegistryServiceMock = {
    findAllForIndexation: jest.fn(),
    countAllUserRegistry: jest.fn(),
    findUsersByNameEmployerOrJobsGroup: jest.fn(),
    initUserRegistryIndex: jest.fn(),
    populateES: jest.fn(),
  };

  beforeEach(async () => {
    const module = await Test.createTestingModule({
      imports: [ConfigurationModule],
      providers: [
        {
          provide: UserRegistryService,
          useValue: userRegistryServiceMock,
        },
        {
          provide: getModelToken('User'),
          useValue: User,
        },
      ],
      controllers: [UsersRegistryController],
    })
      .overrideGuard(JwtAuthGuard)
      .useValue(mockJwtAuthGuard)
      .overrideGuard(RolesGuard)
      .useValue(mockRoleGuard)
      .compile();

    userRegistryController = module.get<UsersRegistryController>(UsersRegistryController);
  });

  it('should be defined', () => {
    expect(userRegistryController).toBeDefined();
  });

  describe('findAll', () => {
    it('should findAll with searchTerm, job and employer', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll(
        { search: 'adm' },
        { page: 1, jobsGroup: ['Technique'], employer: ['Pimms'] }
      );
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with searchTerm, empty job and empty employer', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: 'adm' }, { page: 1, jobsGroup: [], employer: [] });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with searchTerm and no filter arrays', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: 'adm' }, { page: 1 });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with searchTerm and empty job', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: 'adm' }, { page: 1, jobsGroup: [] });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with searchTerm and empty employer', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: 'adm' }, { page: 1, employer: [] });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with no searchTerm and employer filter', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: '' }, { page: 1, employer: ['CAF'] });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with no searchTerm and job filter', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: '' }, { page: 1, jobsGroup: ['Technique'] });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with no searchTerm and filters', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll(
        { search: '' },
        { page: 1, jobsGroup: ['Technique'], employer: ['CAF'] }
      );
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with searchTerm and undefined filters', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue([multipleUsers[0]]);
      const reply = await userRegistryController.findAll({ search: 'adm' });
      expect(reply).toStrictEqual([multipleUsers[0]]);
    });
    it('should findAll with no searchTerm and no filter arrays', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue(multipleUsers);
      const reply = await userRegistryController.findAll({ search: '' }, { page: 1 });
      expect(reply).toBe(multipleUsers);
    });
    it('should findAll with empty search end undefined filters', async () => {
      userRegistryServiceMock.findUsersByNameEmployerOrJobsGroup.mockResolvedValue(multipleUsers);
      const reply = await userRegistryController.findAll({ search: '' });
      expect(reply).toBe(multipleUsers);
    });
  });
  describe('findAllCount', () => {
    it('should findAllCount', async () => {
      userRegistryServiceMock.countAllUserRegistry.mockResolvedValue(10);
      const reply = await userRegistryController.findAllCount();
      expect(reply).toStrictEqual(10);
    });
  });
  describe('resetES', () => {
    it('should reset elastic search indexes', async () => {
      userRegistryServiceMock.initUserRegistryIndex.mockResolvedValue(multipleUsers);
      const reply = await userRegistryController.resetES();
      expect(reply).toStrictEqual(multipleUsers);
    });
  });
});
