import {
  Body,
  Controller,
  Delete,
  Get,
  HttpException,
  HttpStatus,
  Logger,
  Param,
  Post,
  Query,
  Req,
  Request,
  UseGuards,
} from '@nestjs/common';
import { JwtService } from '@nestjs/jwt';
import { ApiBearerAuth, ApiBody, ApiOperation, ApiParam, ApiResponse, ApiTags } from '@nestjs/swagger';
import { DateTime } from 'luxon';
import { Types } from 'mongoose';
import { JwtAuthGuard } from '../../auth/guards/jwt-auth.guard';
import { Structure } from '../../structures/schemas/structure.schema';
import { StructuresService } from '../../structures/services/structures.service';
import { TempUserService } from '../../temp-user/temp-user.service';
import { EmailChangeDto } from '../dto/change-email.dto';
import { PasswordChangeDto } from '../dto/change-password.dto';
import { CreateUserDto } from '../dto/create-user.dto';
import { DescriptionDto } from '../dto/description.dto';
import { ProfileDto } from '../dto/profile.dto';
import { PasswordResetApplyDto } from '../dto/reset-password-apply.dto';
import { PasswordResetCheckDto } from '../dto/reset-password-check.dto';
import { PasswordResetDto } from '../dto/reset-password.dto';
import { UpdateDetailsDto } from '../dto/update-details.dto';
import { IPendingStructureToken } from '../interfaces/pending-structure-token.interface';
import { IUser } from '../interfaces/user.interface';
import { User } from '../schemas/user.schema';
import { EmployerService } from '../services/employer.service';
import { JobsService } from '../services/jobs.service';
import { UsersService } from '../services/users.service';

@ApiTags('users')
@Controller('users')
export class UsersController {
  private readonly logger = new Logger(UsersController.name);
  constructor(
    private usersService: UsersService,
    private structureService: StructuresService,
    private tempUsersService: TempUserService,
    private employerService: EmployerService,
    private jobsService: JobsService,
    private jwtService: JwtService
  ) {}

  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @ApiOperation({ description: 'Get user profile' })
  @ApiResponse({ status: HttpStatus.OK, description: 'Return user profile' })
  @ApiResponse({ status: HttpStatus.UNAUTHORIZED, description: 'User does not have sufficient rights' })
  @Get('profile')
  public getProfile(@Request() req) {
    return req.user;
  }

  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @ApiOperation({ description: 'Find users with specific job criteria and location' })
  @ApiResponse({ status: 200, description: 'Returns filtered list of users' })
  @Get('filteredUsers')
  public async findUsersWithJobWithPersonalOffer(
    @Query('hasPersonalOffer') hasPersonalOffer?: boolean,
    @Query('job') jobs?: string,
    @Query('ctms') ctms?: string,
    @Query('inseeCodes') inseeCodes?: string
  ): Promise<IUser[]> {
    this.logger.log(`findUsersWithJobWithPersonalOffer: ${hasPersonalOffer}; ${jobs}; ${ctms}; ${inseeCodes}`);
    const jobArray = jobs ? jobs.split(',') : undefined;
    const ctmsArray = ctms ? ctms.split(',') : undefined;
    const inseeCodesArray = inseeCodes ? inseeCodes.split(',') : undefined;
    return this.usersService.usersFilteredByJobAndLocation(hasPersonalOffer, jobArray, ctmsArray, inseeCodesArray);
  }

  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @ApiOperation({ description: 'Set user profile with employer and job' })
  @ApiBody({ type: ProfileDto })
  @ApiResponse({ status: HttpStatus.OK, description: 'Return user profile' })
  @ApiResponse({ status: HttpStatus.BAD_REQUEST, description: 'Job does not exist' })
  @ApiResponse({ status: HttpStatus.BAD_REQUEST, description: 'Employer does not exist' })
  @Post('profile')
  public async setProfile(@Request() req, @Body() profile: ProfileDto): Promise<IUser> {
    this.logger.debug(`setProfile`);
    const employerDocument = await this.employerService.findByName(profile.employerName);
    if (!employerDocument) {
      this.logger.warn(`Employer does not exist: ${profile.employerName}`);
      throw new HttpException('Employer does not exist', HttpStatus.BAD_REQUEST);
    }
    const jobDocument = await this.jobsService.findByName(profile.jobName);
    if (!jobDocument) {
      this.logger.warn(`Job does not exist: ${profile.jobName}`);
      throw new HttpException('Job does not exist', HttpStatus.BAD_REQUEST);
    }
    const user = await this.usersService.updateUserProfile(
      req.user._id,
      employerDocument,
      jobDocument,
      profile.withAppointment
    );
    return user;
  }

  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @ApiOperation({ description: 'Updates name, surname and phone number of a user' })
  @ApiBody({ type: UpdateDetailsDto })
  @ApiResponse({ status: HttpStatus.CREATED, description: 'Return user profile' })
  @ApiResponse({ status: HttpStatus.NOT_FOUND, description: 'User not found' })
  @ApiResponse({ status: HttpStatus.BAD_REQUEST, description: 'Missing parameter' })
  @Post('details')
  public async updateDetails(@Request() req, @Body() details: UpdateDetailsDto): Promise<User | HttpException> {
    this.logger.debug('updateProfile details');
    return this.usersService.updateUserDetails(req.user._id, details);
  }

  @Post()
  @ApiResponse({ status: 201, description: 'User created' })
  public async create(@Body() createUserDto: CreateUserDto) {
    this.logger.debug('create');
    const user = await this.usersService.create(createUserDto);
    // Remove temp user if exist
    const tempUser = await this.tempUsersService.findOne(createUserDto.email);
    if (tempUser) {
      this.tempUsersService.delete(createUserDto.email);
    }
    return user;
  }

  @Post('verify/:id')
  @ApiParam({ name: 'id', type: String, required: true })
  @ApiResponse({ status: 201, description: 'User verified' })
  @ApiResponse({ status: 401, description: "This token doesn't exist or is not associate to this user." })
  public async validateUser(@Param() params, @Query('token') token: string) {
    return this.usersService.validateUser(params.id, token);
  }

  @UseGuards(JwtAuthGuard)
  @Post('change-password')
  @ApiResponse({ status: 201, description: 'Password changed' })
  @ApiResponse({ status: 401, description: 'Invalid password' })
  @ApiResponse({ status: 422, description: 'Weak password' })
  public async changePassword(@Request() req, @Body() passwordChangeDto: PasswordChangeDto) {
    return this.usersService.changeUserPassword(
      req.user._id,
      passwordChangeDto.oldPassword,
      passwordChangeDto.newPassword
    );
  }

  @Post('change-email')
  @ApiResponse({ status: 201, description: 'Email confirmation send' })
  @ApiResponse({ status: 401, description: 'Invalid Email' })
  public async changeEmail(@Body() emailChangeDto: EmailChangeDto) {
    return this.usersService.changeUserEmail(emailChangeDto);
  }

  @UseGuards(JwtAuthGuard)
  @Post('verify-change-email')
  @ApiResponse({ status: 201, description: 'Email changed' })
  @ApiResponse({ status: 401, description: 'Invalid Token' })
  public async verifyAndUpdateEmail(@Query('token') token: string) {
    return this.usersService.verifyAndUpdateUserEmail(token);
  }

  @Post('reset-password')
  @ApiResponse({ status: 200, description: 'Email sent if account exist' })
  public async resetPassword(@Body() passwordReset: PasswordResetDto) {
    return this.usersService.sendResetPasswordEmail(passwordReset.email);
  }

  @Post('reset-password/check')
  @ApiResponse({ status: 200, description: 'Check if token is valid' })
  public async resetPasswordCheck(@Body() passwordResetCheckDto: PasswordResetCheckDto) {
    return this.usersService.checkPasswordResetToken(passwordResetCheckDto.token);
  }

  @Post('reset-password/apply')
  @ApiResponse({ status: 200, description: 'Reset password if token and new password are valid' })
  public async resetPasswordApply(@Body() passwordResetApplyDto: PasswordResetApplyDto) {
    return this.usersService.resetPassword(passwordResetApplyDto.password, passwordResetApplyDto.token);
  }

  @Post('verify-exist-user')
  public async verifyUserExist(@Request() _req, @Body() email: { newMail: string }) {
    return this.usersService.verifyUserExist(email.newMail);
  }

  @Delete()
  @UseGuards(JwtAuthGuard)
  public async delete(@Req() req) {
    const user = await this.usersService.deleteOne(req.user.email);
    user.structuresLink.forEach((structureId) => {
      this.usersService.isStructureClaimed(structureId.toString()).then(async (userFound) => {
        if (!userFound) {
          const structure = await this.structureService.findOne(structureId.toString());
          this.structureService.deleteOne(req.user, structure);
        }
      });
    });
    return user;
  }

  @Get('dataConsentValidation')
  @UseGuards(JwtAuthGuard)
  public async dataConsentValidation(@Request() req): Promise<Structure[]> {
    return this.structureService.getAllDataConsentPendingStructures(req.user);
  }

  @UseGuards(JwtAuthGuard)
  @Get(':id')
  @ApiParam({ name: 'id', type: String, required: true })
  public async getUser(@Param() params): Promise<IUser> {
    const user = await this.usersService.findById(params.id);
    if (!user) {
      throw new HttpException('User does not exist', HttpStatus.NOT_FOUND);
    }
    return user;
  }

  @UseGuards(JwtAuthGuard)
  @Get('permalink/:permalink')
  @ApiParam({ name: 'permalink', type: String, required: true })
  public async getUserByPermalink(@Param('permalink') permalink: string): Promise<IUser> {
    const user = await this.usersService.findByPermalink(permalink);
    if (!user) {
      throw new HttpException('User does not exist', HttpStatus.NOT_FOUND);
    }
    return user;
  }

  @ApiResponse({ status: HttpStatus.CREATED, description: 'Description updated' })
  @ApiResponse({ status: HttpStatus.UNAUTHORIZED, description: 'Unauthorized' })
  @UseGuards(JwtAuthGuard)
  @ApiBearerAuth('JWT')
  @ApiOperation({ description: 'Updates description of a user' })
  @ApiBody({ type: DescriptionDto })
  @Post('description')
  public async updateDescription(@Req() req, @Body() body: DescriptionDto): Promise<User> {
    return await this.usersService.updateDescription(req.user._id, body);
  }

  @Post('join-request/:id')
  @ApiParam({ name: 'id', type: String, required: true })
  public async join(@Param('id') id: string, @Body() user: User): Promise<void> {
    // Get structure name
    const structure = await this.structureService.findOne(id);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get user and add structure to user
    const userFromDb = await this.usersService.findOne(user.email);
    if (!userFromDb) {
      throw new HttpException('Invalid User', HttpStatus.NOT_FOUND);
    }
    // See structure to pending
    const pendingStructures = await this.usersService.updatePendingStructureLinked(
      userFromDb.email,
      id,
      structure.structureName
    );
    const token = pendingStructures
      .filter((pending) => pending.id.equals(structure.id))
      .map((pending) => pending.token);

    // Send structure owners an email
    this.structureService.sendStructureJoinRequest(userFromDb, structure, token[0]);
  }

  /** add token in route + add infos in token */
  @Get('join-validate/:token/:status')
  @ApiParam({ name: 'token', type: String, required: true })
  @ApiParam({ name: 'status', type: String, required: true })
  public async joinValidation(
    @Param('token') token: string,
    @Param('status') status: string
  ): Promise<{ structure: Structure; username: string }> {
    const decoded = this.jwtService.decode(token) as IPendingStructureToken;
    const today = DateTime.local().setZone('utc', { keepLocalTime: true });

    if (!token || !status) {
      throw new HttpException('Wrong parameters', HttpStatus.NOT_FOUND);
    }
    if (DateTime.fromISO(decoded.expiresAt) < today) {
      throw new HttpException('Expired or invalid token', HttpStatus.FORBIDDEN);
    }
    // Get structure name
    const structure = await this.structureService.findOne(decoded.idStructure);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get user and add pending structure
    const userFromDb = await this.usersService.findById(decoded.userId);
    if (!userFromDb) {
      throw new HttpException('Invalid User', HttpStatus.NOT_FOUND);
    }
    if (
      !userFromDb.pendingStructuresLink
        .map((pending) => pending.id)
        .some((id) => id.equals(new Types.ObjectId(decoded.idStructure)))
    ) {
      throw new HttpException('User not linked to structure', HttpStatus.NOT_FOUND);
    }
    if (status == 'true') {
      // Accept
      await this.usersService.updateStructureLinked(userFromDb.email, decoded.idStructure);
      await this.usersService.removeFromPendingStructureLinked(userFromDb.email, decoded.idStructure);
    } else {
      // Refuse
      await this.usersService.removeFromPendingStructureLinked(userFromDb.email, decoded.idStructure);
    }
    await this.usersService.sendStructureClaimApproval(userFromDb.email, structure.structureName, status == 'true');

    return { structure, username: `${userFromDb.name} ${userFromDb.surname}` };
  }

  /** Cancel a user's join request */
  @Get('join-cancel/:idStructure/:idUser')
  @ApiParam({ name: 'idStructure', type: String, required: true })
  @ApiParam({ name: 'idUser', type: String, required: true })
  public async joinCancel(@Param('idStructure') idStructure: string, @Param('idUser') idUser: string): Promise<void> {
    // Get structure name
    const structure = await this.structureService.findOne(idStructure);
    if (!structure) {
      throw new HttpException('Invalid Structure', HttpStatus.NOT_FOUND);
    }
    // Get user and add pending structure
    const userFromDb = await this.usersService.findById(idUser);
    if (!userFromDb) {
      throw new HttpException('Invalid User', HttpStatus.NOT_FOUND);
    }
    if (
      !userFromDb.pendingStructuresLink
        .map((pending) => pending.id)
        .some((id) => id.equals(new Types.ObjectId(idStructure)))
    ) {
      throw new HttpException('This structure is in pending state', HttpStatus.NOT_FOUND);
    }
    await this.usersService.removeFromPendingStructureLinked(userFromDb.email, idStructure);
  }
}
