import { HttpModule } from '@nestjs/axios';
import { HttpStatus } from '@nestjs/common';
import { getModelToken } from '@nestjs/mongoose';
import { Test } from '@nestjs/testing';
import * as ejs from 'ejs';
import { Types } from 'mongoose';
import { ConfigurationModule } from '../../configuration/configuration.module';
import { MailerService } from '../../mailer/mailer.service';
import { CreateJobDto } from '../dto/create-job.dto';
import { JobDocument } from '../schemas/job.schema';
import { JobsService } from './jobs.service';
import { JobsGroupsService } from './jobsGroups.service';
import { UsersService } from './users.service';
import { UpdateJobDto } from '../dto/update-job.dto';

const ejsSpy = jest.spyOn(ejs, 'renderFile');

const mockSingleAdminResponse = [
  {
    _id: '6231aefe76598527c8d0b5bc',
    validationToken:
      'cf1c74c22cedb6b575945098db42d2f493fb759c9142c6aff7980f252886f36ee086574ee99a06bc99119079257116c959c8ec870949cebdef2b293666dbca42',
    emailVerified: true,
    email: 'admin@admin.com',
    password: '$2a$12$vLQjJ9zAWyUwiFLeQDa6w.XzrlgPBhw.2GWrjog/yuEjIaZnQwmZu',
    role: 0,
    name: 'admin',
    surname: 'admin',
    personalOffers: [],
  },
];

describe('JobsService', () => {
  let jobsService: JobsService;

  const mockJobModel = {
    create: jest.fn(),
    find: jest.fn().mockReturnThis(),
    populate: jest.fn().mockReturnThis(),
    collation: jest.fn().mockReturnThis(),
    sort: jest.fn().mockReturnThis(),
    exec: jest.fn(),
    findOne: jest.fn().mockReturnThis(),
    findById: jest.fn().mockReturnThis(),
    deleteOne: jest.fn().mockReturnThis(),
  };

  const mockJobsGroupModel = {
    find: jest.fn().mockReturnThis(),
  };

  const mockUserModel = {
    find: jest.fn(),
    findOne: jest.fn(),
    findById: jest.fn(),
  };
  const mockUserService = {
    replaceJobs: jest.fn(),
    getAdmins: jest.fn(),
  };
  const mockMailService = {
    send: jest.fn(),
    getTemplateLocation: jest.fn().mockReturnValue('location'),
    loadJsonConfig: jest.fn().mockReturnValue({ subject: 'subject' }),
    config: {
      templates: {
        adminJobCreate: {
          ejs: 'adminJobCreate.ejs',
          json: 'adminJobCreate.json',
        },
      },
    },
  };

  beforeEach(async () => {
    const module = await Test.createTestingModule({
      imports: [ConfigurationModule, HttpModule],
      providers: [
        JobsService,
        {
          provide: getModelToken('Job'),
          useValue: mockJobModel,
        },
        {
          provide: JobsGroupsService,
          useValue: mockJobsGroupModel,
        },
        {
          provide: UsersService,
          useValue: mockUserService,
        },
        {
          provide: getModelToken('User'),
          useValue: mockUserModel,
        },
        { provide: MailerService, useValue: mockMailService },
      ],
    }).compile();
    jobsService = module.get<JobsService>(JobsService);
    ejsSpy.mockResolvedValue('mock');
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(jobsService).toBeDefined();
  });

  const mockCNFSJob = {
    _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
    name: 'CNFS',
    validated: true,
    hasPersonalOffer: true,
    jobsGroup: null,
  } as const;

  describe('findAll', () => {
    it('should findAll validated jobs', async () => {
      const results = [mockCNFSJob];
      mockJobModel.exec.mockResolvedValueOnce(results);
      const reply = await jobsService.findAll();
      expect(reply.length).toBe(1);
    });
  });

  it('findUnvalidated', async () => {
    mockJobModel.exec.mockResolvedValueOnce([
      {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'CNFS',
        validated: false,
        hasPersonalOffer: true,
        jobsGroup: null,
      },
    ]);
    const reply = await jobsService.findAllUnvalidated();
    expect(reply[0].validated).toBeFalsy();
  });

  it('findByName', async () => {
    mockJobModel.exec.mockResolvedValueOnce(mockCNFSJob);
    const reply = await jobsService.findByName('CNFS');
    expect(reply).toBeTruthy();
  });
  it('findOne', async () => {
    mockJobModel.exec.mockResolvedValueOnce(mockCNFSJob);
    const reply = await jobsService.findOne('6231aefe76598527c8d0b5bc');
    expect(reply).toBeTruthy();
  });

  describe('createJob', () => {
    it('create', async () => {
      mockJobModel.create.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Dev',
        validated: false,
        hasPersonalOffer: false,
        jobsGroup: null,
      });
      const createJob: CreateJobDto = {
        name: 'Dev',
        hasPersonalOffer: false,
      };
      mockUserService.getAdmins.mockResolvedValueOnce(mockSingleAdminResponse);
      const reply = await jobsService.create(createJob);
      expect(reply).toBeTruthy();
    });
    it('create job and send admin notification', async () => {
      const sendAdminMailSpy = jest.spyOn(jobsService, 'sendAdminCreateNotification');
      mockJobModel.create.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Dev',
        validated: false,
        hasPersonalOffer: false,
        jobsGroup: null,
      });
      const createJob: CreateJobDto = {
        name: 'Dev',
        hasPersonalOffer: false,
      };
      mockUserService.getAdmins.mockResolvedValueOnce(mockSingleAdminResponse);
      const reply = await jobsService.create(createJob, false, false, true);
      expect(reply).toBeTruthy();
      expect(sendAdminMailSpy).toHaveBeenCalled();
    });
  });

  describe('sendAdminCreateNotification', () => {
    it('should sendAdminCreateNotification', async () => {
      const job = {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'Metro',
        validated: true,
      };
      mockUserService.getAdmins.mockResolvedValueOnce(mockSingleAdminResponse);
      await jobsService.sendAdminCreateNotification(job as JobDocument, 'adminJobCreate.ejs', 'adminJobCreate.json');
      expect(mockMailService.send).toHaveBeenCalledTimes(1);
    });
  });

  describe('validate', () => {
    it('should validate job', async () => {
      mockJobModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Dev',
        validated: false,
        save: jest.fn().mockResolvedValueOnce(null),
      });
      const job = await jobsService.validate('623aed68c5d45b6fbbaa7e60');
      expect(job.validated).toBe(true);
    });
    it('should throw exception', async () => {
      mockJobModel.exec.mockResolvedValueOnce(null);
      try {
        await jobsService.validate('623aed68c5d45b6fbbaa7e60');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Cannot validate job. It might have been already validate');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
  });

  describe('mergeJob', () => {
    it('should delete source job', async () => {
      const result = {
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e61'),
        name: 'Dev 2',
        validated: true,
        hasPersonalOffer: false,
        jobsGroup: null,
      };
      mockJobModel.exec
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
          name: 'Dev',
          validated: false,
          hasPersonalOffer: false,
          jobsGroup: null,
        })
        .mockResolvedValueOnce(result);

      const reply = await jobsService.mergeJob({
        sourceJobId: '623aed68c5d45b6fbbaa7e60',
        targetJobId: '623aed68c5d45b6fbbaa7e61',
      });
      expect(reply).toBe(result);
    });
    it('should throw error if target job is not validated', async () => {
      mockJobModel.exec
        .mockResolvedValueOnce({
          _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
          name: 'Dev',
          validated: false,
          hasPersonalOffer: false,
          jobsGroup: null,
        })
        .mockResolvedValueOnce(null);
      try {
        await jobsService.mergeJob({
          sourceJobId: '623aed68c5d45b6fbbaa7e60',
          targetJobId: '623aed68c5d45b6fbbaa7e61',
        });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Cannot operate on job.');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
    it('should throw error if no source or target job', async () => {
      mockJobModel.exec.mockResolvedValueOnce(null).mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Dev',
        validated: false,
        hasPersonalOffer: false,
        jobsGroup: null,
      });
      try {
        await jobsService.mergeJob({
          sourceJobId: '623aed68c5d45b6fbbaa7e60',
          targetJobId: '623aed68c5d45b6fbbaa7e61',
        });
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Cannot operate on job.');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
  });

  it('deleteInvalidJob', async () => {
    mockJobModel.exec.mockResolvedValueOnce(null);
    const spyer = jest.spyOn(jobsService, 'deleteInvalidJob');
    await jobsService.deleteInvalidJob('623aed68c5d45b6fbbaa7e60');
    expect(spyer.mock.calls.length).toEqual(1);
  });

  describe('deleteOneId', () => {
    it('should delete', async () => {
      mockJobModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'Dev',
        validated: false,
        deleteOne: jest.fn().mockResolvedValueOnce('toto'),
      });
      const reply = await jobsService.deleteOneId('623aed68c5d45b6fbbaa7e60');
      expect(reply).toBe('toto');
    });
    it('should not be able to delete', async () => {
      mockJobModel.exec.mockResolvedValueOnce(null);
      try {
        await jobsService.deleteOneId('623aed68c5d45b6fbbaa7e60');
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Invalid job id');
        expect(e.status).toBe(HttpStatus.BAD_REQUEST);
      }
    });
  });

  describe('update', () => {
    const newJobDto: UpdateJobDto = {
      name: 'CNFS',
      hasPersonalOffer: true,
      jobsGroup: null,
    };
    it('should not be able to update', async () => {
      mockJobModel.exec.mockResolvedValueOnce(null);
      try {
        await jobsService.update('623aed68c5d45b6fbbaa7e60', newJobDto);
        expect(true).toBe(false);
      } catch (e) {
        expect(e.message).toBe('Cannot edit job. It was not found in database.');
        expect(e.status).toBe(HttpStatus.NOT_FOUND);
      }
    });
    it('should update', async () => {
      const newJob = {
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'CNFS',
        validated: true,
        hasPersonalOffer: true,
        jobsGroup: null,
      };
      mockJobModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('6231aefe76598527c8d0b5bc'),
        name: 'CNFS old',
        validated: true,
        hasPersonalOffer: false,
        jobsGroup: null,
        save: jest.fn().mockResolvedValueOnce(newJob),
      });
      const reply = await jobsService.update('623aed68c5d45b6fbbaa7e60', newJobDto);
      expect(reply.name).toBe('CNFS');
      expect(reply.hasPersonalOffer).toBe(true);
    });
    it('should update and apply jobsGroup', async () => {
      const jobsGroupID = '65fd51d64bce0eed5ea130a0';
      const jobsGroup = {
        _id: jobsGroupID,
        name: 'Coordination',
      };
      const newJobWithGroup: UpdateJobDto = {
        name: 'CNFS',
        hasPersonalOffer: true,
        jobsGroup: jobsGroup as any,
      };
      mockJobModel.exec.mockResolvedValueOnce({
        _id: new Types.ObjectId('623aed68c5d45b6fbbaa7e60'),
        name: 'CNFS',
        validated: true,
        hasPersonalOffer: true,
        jobsGroup: jobsGroup,
        save: jest.fn().mockResolvedValueOnce(newJobWithGroup),
      });
      const reply = await jobsService.update('623aed68c5d45b6fbbaa7e60', newJobWithGroup);
      expect(reply.jobsGroup).toEqual(new Types.ObjectId(jobsGroupID));
    });
  });

  it('should call getJobsGroupsAttachedJobs', async () => {
    await jobsService.getJobsGroupsAttachedJobs('623aed68c5d45b6fbbaa7e60');
    expect(mockJobModel.find).toHaveBeenCalledTimes(1);
  });

  describe('isConseillerNumerique', () => {
    it('should be ConseillerNumerique', async () => {
      const response = jobsService.isConseillerNumerique(new Types.ObjectId('63beead900d7a73701776fd9'));
      expect(response).toBe(true);
    });
    it('should NOT be ConseillerNumerique', async () => {
      const response = jobsService.isConseillerNumerique(new Types.ObjectId('03beead900d7a73701776fd0'));
      expect(response).toBe(false);
    });
  });
});
