import { HttpException, HttpStatus, Inject, Injectable, Logger, forwardRef } from '@nestjs/common';
import { InjectModel } from '@nestjs/mongoose';
import * as ejs from 'ejs';
import * as mongodb from 'mongodb';
import { Model, Types } from 'mongoose';
import { MergeJobDto } from '../../admin/dto/merge-job.dto';
import { MailerService } from '../../mailer/mailer.service';
import { CreateJobDto } from '../dto/create-job.dto';
import { UpdateJobDto } from '../dto/update-job.dto';
import { IJob } from '../interfaces/job.interface';
import { Job, JobDocument } from '../schemas/job.schema';
import { UsersService } from './users.service';

@Injectable()
export class JobsService {
  // CNFS job ids ('Conseiller numérique' or 'Conseillère numérique')
  public readonly CNFSJobsId = [
    new Types.ObjectId('63beead900d7a73701776fd9'),
    new Types.ObjectId('646f2cb009f08f048d40f689'),
  ];
  private readonly logger = new Logger(JobsService.name);

  constructor(
    @InjectModel(Job.name) private jobModel: Model<JobDocument>,
    @Inject(forwardRef(() => UsersService))
    private readonly userService: UsersService,
    private readonly mailerService: MailerService
  ) {}

  public async findAll(searchString = ''): Promise<IJob[]> {
    this.logger.debug('findAll ' + (searchString ? `with query ${searchString}` : 'without query'));
    return this.jobModel
      .find({ validated: true, name: { $regex: searchString, $options: 'i' } })
      .populate('jobsGroup')
      .collation({ locale: 'fr' })
      .sort({ name: 1 })
      .exec();
  }

  public async findByName(name: string): Promise<JobDocument> {
    this.logger.debug('findByName');
    return this.jobModel.findOne({ name }).exec();
  }

  public async findOne(idParam: string): Promise<JobDocument> {
    this.logger.debug('findOne');
    return this.jobModel.findById(new Types.ObjectId(idParam)).exec();
  }

  public async findAllUnvalidated(): Promise<IJob[]> {
    this.logger.debug('findAllUnvalidated');
    return this.jobModel.find({ validated: false }).exec();
  }

  public async create(
    job: CreateJobDto,
    validated = false,
    hasPersonalOffer = false,
    sendNotification = true
  ): Promise<Job> {
    this.logger.debug(`createJob: ${job.name} | notificationSending ${sendNotification}`);
    return this.jobModel.create({ name: job.name, validated, hasPersonalOffer }).then(async (postCreate) => {
      if (sendNotification) {
        this.sendAdminCreateNotification(
          postCreate,
          this.mailerService.config.templates.adminJobCreate.ejs,
          this.mailerService.config.templates.adminJobCreate.json
        );
      }
      return postCreate;
    });
  }

  // TODO templateLocation and jsonConfigLocation should not be passed as arguments
  public async sendAdminCreateNotification(
    job: JobDocument,
    templateLocation: string,
    jsonConfigLocation: string
  ): Promise<void> {
    const config = this.mailerService.config;
    const ejsPath = this.mailerService.getTemplateLocation(templateLocation);
    const jsonConfig = this.mailerService.loadJsonConfig(jsonConfigLocation);
    const html = await ejs.renderFile(ejsPath, {
      config,
      jobName: job ? job.name : '',
    });
    const admins = await this.userService.getAdmins();
    admins.forEach((admin) => {
      this.mailerService.send(admin.email, jsonConfig.subject, html);
    });
  }

  public async validate(jobId: string): Promise<Job> {
    this.logger.debug(`validateJob: ${jobId}`);
    const job = await this.jobModel.findById(new Types.ObjectId(jobId)).exec();
    if (!job) {
      throw new HttpException('Cannot validate job. It might have been already validate', HttpStatus.NOT_FOUND);
    }
    job.validated = true;
    job.save();
    return job;
  }

  public async mergeJob({ sourceJobId, targetJobId }: MergeJobDto): Promise<Job | HttpException> {
    this.logger.debug(`mergeJob: ${sourceJobId} into ${targetJobId}`);
    const sourceJob = await this.jobModel.findById(new Types.ObjectId(sourceJobId)).exec();
    const targetJob = await this.jobModel.findById(new Types.ObjectId(targetJobId)).exec();
    if (targetJob && sourceJob) {
      this.logger.debug(`Both jobs : ${JSON.stringify(sourceJob)}, ${JSON.stringify(targetJob)}`);
      await this.userService.replaceJobs(sourceJob, targetJob);
      if (!sourceJob.validated) {
        await this.deleteInvalidJob(sourceJobId);
      }
      return targetJob;
    } else {
      throw new HttpException('Cannot operate on job.', HttpStatus.NOT_FOUND);
    }
  }

  public async deleteInvalidJob(id: string): Promise<mongodb.DeleteResult> {
    this.logger.debug(`deleteInvalidJob: ${id}`);
    return this.jobModel.deleteOne({ _id: id }).exec();
  }

  public async deleteOneId(id: string): Promise<Job> {
    const job = await this.jobModel.findOne({ _id: id }).exec();
    if (!job) {
      throw new HttpException('Invalid job id', HttpStatus.BAD_REQUEST);
    }
    return job.deleteOne();
  }

  public async update(jobId: string, newJob: UpdateJobDto): Promise<Job> {
    this.logger.debug(`editJob: ${jobId}`);
    const job = await this.jobModel.findById(new Types.ObjectId(jobId)).exec();
    if (!job) {
      throw new HttpException('Cannot edit job. It was not found in database.', HttpStatus.NOT_FOUND);
    }
    job.name = newJob.name;
    job.hasPersonalOffer = newJob.hasPersonalOffer;
    if (newJob.jobsGroup) {
      job.jobsGroup = new Types.ObjectId(newJob.jobsGroup._id);
    } else {
      job.jobsGroup = undefined;
    }
    job.save();
    return job;
  }

  public getJobsGroupsAttachedJobs(jobsGroupId: string): Promise<JobDocument[]> {
    this.logger.debug(`getJobsGroupsAttachedJobs: ${jobsGroupId}`);
    return this.jobModel.find({ jobsGroup: jobsGroupId }).exec();
  }

  public isConseillerNumerique(jobId: Types.ObjectId) {
    // If job is 'Conseiller numérique' or 'Conseillère numérique'
    return this.CNFSJobsId.some((cnfsId) => cnfsId.equals(jobId));
  }
}
