import { IndicesCreateResponse, IndicesRefreshResponse, WriteResponseBase } from '@elastic/elasticsearch/lib/api/types';
import { Injectable, Logger } from '@nestjs/common';
import { ElasticsearchService } from '@nestjs/elasticsearch';
import { es_settings_homemade_french, escapeElasticsearchQuery } from '../../shared/utils';
import { UserRegistrySearchBody } from '../interfaces/userRegistry-search-body.interface';
import { IUserRegistry } from '../interfaces/userRegistry.interface';

@Injectable()
export class UserRegistrySearchService {
  private index = 'user';
  private readonly logger = new Logger(UserRegistrySearchService.name);

  constructor(private readonly elasticsearchService: ElasticsearchService) {}

  public async indexUserRegistry(user: IUserRegistry): Promise<IUserRegistry> {
    this.logger.debug(`indexUser ${user.name} ${user.surname}`);
    this.elasticsearchService.index<UserRegistrySearchBody>({
      index: this.index,
      id: user._id,
      document: {
        name: user.name,
        id: user._id,
        surname: user.surname,
        job: user.job,
        employer: user.employer,
      },
    });
    return user;
  }

  public async createUserRegistryIndex(): Promise<IndicesCreateResponse> {
    this.logger.debug('createUserRegistryIndex');
    return this.elasticsearchService.indices.create({
      index: this.index,
      mappings: {
        dynamic_templates: [
          {
            strings: {
              match_mapping_type: 'string',
              mapping: {
                type: 'text',
                analyzer: 'homemade_french',
                search_analyzer: 'homemade_french_stopless',
              },
            },
          },
        ],
      },
      settings: es_settings_homemade_french,
    });
  }

  public async dropIndex() {
    this.logger.debug('dropIndex');
    const foundIndexes = await this.elasticsearchService.indices.exists({
      index: this.index,
    });
    if (foundIndexes) {
      return this.elasticsearchService.indices.delete({
        index: this.index,
      });
    }
  }

  public async deleteIndex(user: IUserRegistry): Promise<WriteResponseBase> {
    this.logger.debug('deleteIndex');
    return this.elasticsearchService.delete({
      index: this.index,
      id: user._id,
    });
  }

  public async refreshIndexUserRegistry(): Promise<IndicesRefreshResponse> {
    this.logger.debug('refreshIndexUserRegistry');
    return this.elasticsearchService.indices.refresh({
      index: this.index,
    });
  }

  public async search(searchString: string): Promise<UserRegistrySearchBody[]> {
    this.logger.debug(`search user with query: ${searchString}`);
    searchString = searchString ? searchString + '*' : '*';
    const { hits } = await this.elasticsearchService.search<UserRegistrySearchBody>({
      index: this.index,
      from: 0,
      size: 5000,
      query: {
        query_string: {
          analyze_wildcard: true,
          query: escapeElasticsearchQuery(searchString),
          fields: ['name', 'surname'],
          fuzziness: 'AUTO',
        },
      },
    });
    const sortedHits = hits.hits.filter(function (elem) {
      return elem._score >= hits.max_score / 1.5;
    });
    return sortedHits.map((item) => item._source);
  }

  public async update(user: IUserRegistry) {
    this.logger.debug(`Updates user : ${user._id}`);
    return this.elasticsearchService.update<UserRegistrySearchBody>({
      index: this.index,
      id: user._id,
      doc: {
        name: user.name,
        surname: user.surname,
        job: user.job ?? null,
        employer: user.employer ?? null,
      },
    });
  }
}
